import numpy as np
from scipy.interpolate import splprep, splev
from numpy.typing import ArrayLike
from typing import Optional


def spline_trajectory(
    t: ArrayLike,
    control_points: ArrayLike,
    *,
    t_control: Optional[ArrayLike] = None,
    degree: int = 3,
    t_min: float = 0,
    t_max: float = 1,
    derivative: int = 0,
) -> np.ndarray:
    """Evaluate the trajectory given by control_points at t using B-spline
    interpolation.

    ``spline_trajectory`` constructs a ``degree``-times differentiable
    trajectory using the given control points and then evaluates the resulting
    trajectory at ``t``. It does so using B-splines. By default, control points
    are spaced out evenly in the interval ``[t_min, t_max]`` where ``t=t_min``
    results in ``control_points[0]`` and ``t=t_max`` results in
    ``control_poins[-1]``. Alternatively, the spacing of control points can be
    set manually by specifying ``t_control``, which implicitly specifies
    ``t_min`` and ``t_max``.

    Parameters
    ----------
    t : np.ndarray
        An array containing positions at which to evaluate the trajectory.
        Elements of ``t`` must be within ``[t_min, t_max]``.
    control_points : np.ndarray
        A batch of control points used to construct the trajectory. The first
        dimension of the array is interpreted as batch dimension and the
        remaining dimensions are used to interpolate between. By default,
        control points are equally spaced within ``[t_min, t_max]`` unless
        ``t_control`` is given explicitly.
    t_control : np.ndarray, None
        A sequence of strictly increasing floats determining the position of the
        control points along the trajectory. None by default, which results in
        an equidistant spacing of points.
    degree : int
        The degree of the spline; uneven numbers are preferred. The resulting
        spline is k times continously differentiable.
    t_min : float
        Minimum value of the trajectories parametrization. Must be smaller than
        ``t_max``. If ``t_control`` is set, this value is ignored in favor of
        ``t_min=t_control[0]``
    t_max : float
        Maximum value of the trajectories parametrization. Must be larger than
        ``t_min``. If ``t_control`` is set, this value is ignored in favor of
        ``t_max=t_control[-1]``.
    derivative : int
        The derivative of the interpolated trajectory to compute. For example,
        ``derivative=2`` differentiates the trajectory twice with respect to
        ``t`` and then evaluates the derivative at the given ``t``.

    Returns
    -------
    position : np.ndarray
        The value of the trajectory at ``t``.

    Notes
    -----
    The dimension of the space embedding the trajectory must be less than 12,
    i.e. ``<= 11``, due to limitations in scipy. If more dimensions are needed,
    please open an issue; a workaround is to split the trajectory into chunks
    of less than 11 dimensions each.

    Repeated evaluation of single points on the trajectory, i.e. repeatedly
    calling this function with scalar ``t``, is possible, but will repeatedly
    reconstruct the trajectory, which can lead to unnecessary slowdown. For
    better performance, it is preferred to use an array-like ``t``.

    Examples
    --------

    .. plot::
        :include-source:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from skbot.trajectory import spline_trajectory
        >>> t1 = np.linspace(0, 2*np.pi, 10)
        >>> control_points = np.stack((np.cos(t1), np.sin(t1)), axis=1)
        >>> t2 = np.linspace(0, 2*np.pi, 100)
        >>> trajectory = spline_trajectory(t2, control_points, t_min=0, t_max=2*np.pi)
        >>> fig, ax = plt.subplots()
        >>> ax.plot(trajectory[:,0], trajectory[:,1], control_points[:,0], control_points[:,1], 'o')
        >>> fig.legend(('Trajectory', 'Control Points'))
        >>> plt.show()

    """
    t = np.asarray(t)
    control_points = np.asarray(control_points)

    if t_control is None:
        t_control = np.linspace(t_min, t_max, len(control_points), dtype=np.float_)
    else:
        t_control = np.asarray(t_control)
        t_min = t_control[0]
        t_max = t_control[-1]

    tck, u = splprep(control_points.T, u=t_control, s=0, ub=t_min, ue=t_max, k=degree)
    return np.stack(splev(t, tck, der=derivative, ext=2), axis=-1)



from xml.etree import ElementTree
from scipy.spatial.transform import Rotation
from typing import Dict, List, Tuple
import numpy as np

from ... import transform as rtf


def create_frame_graph(urdf: str) -> Tuple[Dict[str, rtf.Frame], Dict[str, rtf.Link]]:
    """Create a frame graph from a URDF string.

    Parameters
    ----------
    urdf: TextIO
        A text buffer containing the URDF XML.

    Returns
    -------
    frames : Dict[str, Frames]
        A dict of frames (links and joints) contained in the file.
    links : Dict[str, Frames]
        A dict of links in the graph. Names are chosen based on joint names.

    See Also
    --------
    :mod:`skbot.transform`

    Notes
    -----
    ``frames[jointName]`` will return the joint's frame that is attached to the
    parent.

    """

    tree = ElementTree.fromstring(urdf)

    frames = dict()
    links = dict()
    links_to_process: List[ElementTree.Element] = list()

    for child in tree:
        if child.tag == "link":
            frames[child.attrib["name"]] = rtf.Frame(3, name=child.attrib["name"])
        elif child.tag == "joint":
            frames[child.attrib["name"]] = rtf.Frame(3, name=child.attrib["name"])
            links_to_process.append(child)

    for link in links_to_process:
        frame_joint = frames[link.attrib["name"]]
        joint_type = link.attrib["type"]

        frame_offset = np.zeros(3, dtype=np.float_)
        frame_rotation = np.zeros(3, dtype=np.float_)
        for child in link:
            if child.tag == "parent":
                frame_parent = frames[child.attrib["link"]]
            elif child.tag == "child":
                frame_child = frames[child.attrib["link"]]
            elif child.tag == "origin":
                try:
                    frame_offset = child.attrib["xyz"]
                    frame_offset = -np.array(frame_offset.split(" "), dtype=np.float_)
                except KeyError:
                    frame_offset = np.zeros(3, dtype=np.float_)

                try:
                    frame_rotation = child.attrib["rpy"]
                    frame_rotation = np.array(
                        frame_rotation.split(" "), dtype=np.float_
                    )
                except KeyError:
                    frame_rotation = np.zeros(3, dtype=np.float_)
            elif child.tag == "axis":
                axis = child.attrib["xyz"]
                axis = np.array(axis.split(" "), dtype=np.float_)

        # link parent -> joint
        rotation = Rotation.from_euler("xyz", frame_rotation)
        if rotation.magnitude() > 0:
            intermediate_frame = rtf.EulerRotation("xyz", frame_rotation)(frame_parent)
            rtf.affine.Translation(frame_offset)(intermediate_frame, frame_joint)
        else:
            rtf.affine.Translation(frame_offset)(frame_parent, frame_joint)

        # link joint -> child
        if joint_type == "fixed":
            frame_link = rtf.affine.Translation((0, 0, 0))
        elif joint_type == "revolute":
            frame_link = rtf.RotvecRotation(axis, angle=0)
        elif joint_type == "prismatic":
            frame_link = rtf.affine.Translation(-axis, amount=0)
        else:
            raise ValueError(f"Unsupported Joint type {joint_type}")

        frame_link(frame_joint, frame_child)

        links[link.attrib["name"]] = frame_link

    return frames, links



from dataclasses import dataclass, field
from typing import List, Optional
from .collision import Collision
from .light import Light
from .material import Material
from .sensor import Sensor
from .visual import Visual

__NAMESPACE__ = "sdformat/v1.7/link.xsd"


@dataclass
class Link:
    """A physical link with inertia, collision, and visual properties.

    A link must be a child of a model, and any number of links may exist
    in a model.

    Parameters
    ----------
    gravity: If true, the link is affected by gravity.
    enable_wind: If true, the link is affected by the wind.
    self_collide: If true, the link can collide with other links in the
        model. Two links within a model will collide if
        link1.self_collide OR link2.self_collide. Links connected by a
        joint will never collide.
    kinematic: If true, the link is kinematic only
    must_be_base_link: If true, the link will have 6DOF and be a direct
        child of world.
    velocity_decay: Exponential damping of the link's velocity.
    pose: A position(x,y,z) and orientation(roll, pitch yaw) with
        respect to the frame named in the relative_to attribute.
    inertial: The inertial properties of the link.
    collision: The collision properties of a link. Note that this can be
        different from the visual properties of a link, for example,
        simpler collision models are often used to reduce computation
        time.
    visual: The visual properties of the link. This element specifies
        the shape of the object (box, cylinder, etc.) for visualization
        purposes.
    sensor: The sensor tag describes the type and properties of a
        sensor.
    projector:
    audio_sink: An audio sink.
    audio_source: An audio source.
    battery: Description of a battery.
    light: The light element describes a light source.
    particle_emitter: A particle emitter that can be used to describe
        fog, smoke, and dust.
    name: A unique name for the link within the scope of the model.
    """

    class Meta:
        name = "link"

    gravity: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    enable_wind: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    self_collide: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    kinematic: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    must_be_base_link: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    velocity_decay: Optional["Link.VelocityDecay"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    pose: Optional["Link.Pose"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    inertial: Optional["Link.Inertial"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    collision: List[Collision] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    visual: List[Visual] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    sensor: List[Sensor] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    projector: Optional["Link.Projector"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    audio_sink: List[str] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    audio_source: List["Link.AudioSource"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    battery: List["Link.Battery"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    light: List[Light] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    particle_emitter: List["Link.ParticleEmitter"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class VelocityDecay:
        """
        Exponential damping of the link's velocity.

        Parameters
        ----------
        linear: Linear damping
        angular: Angular damping
        """

        linear: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        angular: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Pose:
        """
        Parameters
        ----------
        value:
        relative_to: Name of frame relative to which the pose is
            applied.
        """

        value: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        relative_to: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            },
        )

    @dataclass
    class Inertial:
        """
        The inertial properties of the link.

        Parameters
        ----------
        mass: The mass of the link.
        inertia: The 3x3 rotational inertia matrix. Because the
            rotational inertia matrix is symmetric, only 6 above-
            diagonal elements of this matrix are specified here, using
            the attributes ixx, ixy, ixz, iyy, iyz, izz.
        pose: This is the pose of the inertial reference frame, relative
            to the specified reference frame. The origin of the inertial
            reference frame needs to be at the center of gravity. The
            axes of the inertial reference frame do not need to be
            aligned with the principal axes of the inertia.
        """

        mass: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        inertia: Optional["Link.Inertial.Inertia"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        pose: Optional["Link.Inertial.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Inertia:
            """The 3x3 rotational inertia matrix.

            Because the rotational inertia matrix is symmetric, only 6
            above-diagonal elements of this matrix are specified here,
            using the attributes ixx, ixy, ixz, iyy, iyz, izz.
            """

            ixx: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            ixy: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            ixz: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            iyy: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            iyz: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            izz: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            relative_to: Name of frame relative to which the pose is
                applied.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            relative_to: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

    @dataclass
    class Projector:
        """
        Parameters
        ----------
        texture: Texture name
        fov: Field of view
        near_clip: Near clip distance
        far_clip: far clip distance
        pose: A position(x,y,z) and orientation(roll, pitch yaw) with
            respect to the frame named in the relative_to attribute.
        plugin: A plugin is a dynamically loaded chunk of code. It can
            exist as a child of world, model, and sensor.
        name: Name of the projector
        """

        texture: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        fov: float = field(
            default=0.785,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        near_clip: float = field(
            default=0.1,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        far_clip: float = field(
            default=10.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        pose: Optional["Link.Projector.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        plugin: List["Link.Projector.Plugin"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            relative_to: Name of frame relative to which the pose is
                applied.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            relative_to: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

        @dataclass
        class Plugin:
            """A plugin is a dynamically loaded chunk of code.

            It can exist as a child of world, model, and sensor.

            Parameters
            ----------
            any_element: This is a special element that should not be
                specified in an SDFormat file. It automatically copies
                child elements into the SDFormat element so that a
                plugin can access the data.
            name: A unique name for the plugin, scoped to its parent.
            filename: Name of the shared library to load. If the
                filename is not a full path name, the file will be
                searched for in the configuration paths.
            """

            any_element: List[object] = field(
                default_factory=list,
                metadata={
                    "type": "Wildcard",
                    "namespace": "##any",
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )
            filename: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

    @dataclass
    class AudioSource:
        """
        An audio source.

        Parameters
        ----------
        uri: URI of the audio media.
        pitch: Pitch for the audio media, in Hz
        gain: Gain for the audio media, in dB.
        contact: List of collision objects that will trigger audio
            playback.
        loop: True to make the audio source loop playback.
        pose: A position(x,y,z) and orientation(roll, pitch yaw) with
            respect to the frame named in the relative_to attribute.
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        pitch: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        gain: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        contact: Optional["Link.AudioSource.Contact"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        loop: bool = field(
            default=False,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        pose: Optional["Link.AudioSource.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Contact:
            """
            List of collision objects that will trigger audio playback.

            Parameters
            ----------
            collision: Name of child collision element that will trigger
                audio playback.
            """

            collision: List[str] = field(
                default_factory=list,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "min_occurs": 1,
                },
            )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            relative_to: Name of frame relative to which the pose is
                applied.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            relative_to: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

    @dataclass
    class Battery:
        """
        Description of a battery.

        Parameters
        ----------
        voltage: Initial voltage in volts.
        name: Unique name for the battery.
        """

        voltage: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class ParticleEmitter:
        """
        A particle emitter that can be used to describe fog, smoke, and dust.

        Parameters
        ----------
        emitting: True indicates that the particle emitter should
            generate particles when loaded
        duration: The number of seconds the emitter is active. A value
            less than or equal to zero means infinite duration.
        size: The size of the emitter where the particles are sampled.
            Default value is (1, 1, 1).     Note that the interpretation
            of the emitter area varies     depending on the emmiter
            type:       - point: The area is ignored.       - box: The
            area is interpreted as width X height X depth.       -
            cylinder: The area is interpreted as the bounding box of the
            cylinder. The cylinder is oriented along the Z-axis.       -
            ellipsoid: The area is interpreted as the bounding box of an
            ellipsoid shaped area, i.e. a sphere or
            squashed-sphere area. The parameters are again
            identical to EM_BOX, except that the dimensions
            describe the widest points along each of the axes.
        particle_size: The particle dimensions (width, height, depth).
        lifetime: The number of seconds each particle will ’live’ for
            before being destroyed. This value must be greater than
            zero.
        rate: The number of particles per second that should be emitted.
        min_velocity: Sets a minimum velocity for each particle (m/s).
        max_velocity: Sets a maximum velocity for each particle (m/s).
        scale_rate: Sets the amount by which to scale the particles in
            both x and y direction per second.
        color_start: Sets the starting color for all particles emitted.
            The actual color will be interpolated between this color
            and the one set under color_end.      Color::White is the
            default color for the particles      unless a specific
            function is used.      To specify a color, RGB values should
            be passed in.      For example, to specify red, a user
            should enter:      &amp;lt;color_start&amp;gt;1 0
            0&amp;lt;/color_start&amp;gt;      Note that this function
            overrides the particle colors set      with
            color_range_image.
        color_end: Sets the end color for all particles emitted.     The
            actual color will be interpolated between this color     and
            the one set under color_start.     Color::White is the
            default color for the particles     unless a specific
            function is used (see color_start for     more information
            about defining custom colors with RGB     values).     Note
            that this function overrides the particle colors set
            with color_range_image.
        color_range_image: Sets the path to the color image used as an
            affector. This affector modifies the color of particles in
            flight. The colors are taken from a specified image file.
            The range of color values begins from the left side of the
            image and moves to the right over the lifetime of the
            particle, therefore only the horizontal dimension of the
            image is used.  Note that this function overrides the
            particle colors set with color_start and color_end.
        topic: Topic used to update particle emitter properties at
            runtime.      The default topic is
            /model/{model_name}/particle_emitter/{emitter_name}
            Note that the emitter id and name may not be changed.
        particle_scatter_ratio: This is used to determine the ratio of
            particles that will be detected     by sensors. Increasing
            the ratio means there is a higher chance of     particles
            reflecting and interfering with depth sensing, making the
            emitter appear more dense. Decreasing the ratio decreases
            the chance     of particles reflecting and interfering with
            depth sensing, making it     appear less dense.
        pose: A position(x,y,z) and orientation(roll, pitch yaw) with
            respect to the frame named in the relative_to attribute.
        material: The material of the visual element.
        name: A unique name for the particle emitter.
        type: The type of a particle emitter. One of "box", "cylinder",
            "ellipsoid", or "point".
        """

        emitting: bool = field(
            default=True,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        duration: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        size: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        particle_size: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        lifetime: float = field(
            default=5.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        rate: float = field(
            default=10.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        min_velocity: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        max_velocity: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale_rate: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        color_start: str = field(
            default="1 1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )
        color_end: str = field(
            default="1 1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )
        color_range_image: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        topic: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        particle_scatter_ratio: float = field(
            default=0.65,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        pose: Optional["Link.ParticleEmitter.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        material: Optional[Material] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        type: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            relative_to: Name of frame relative to which the pose is
                applied.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            relative_to: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )



from dataclasses import dataclass, field
from typing import List, Optional
from .joint import Joint
from .link import Link

__NAMESPACE__ = "sdformat/v1.5/model.xsd"


@dataclass
class Model:
    """
    The model element defines a complete robot or any other physical object.

    Parameters
    ----------
    static: If set to true, the model is immovable. Otherwise the model
        is simulated in the dynamics engine.
    self_collide: If set to true, all links in the model will collide
        with each other (except those connected by a joint). Can be
        overridden by the link or collision element self_collide
        property. Two links within a model will collide if
        link1.self_collide OR link2.self_collide. Links connected by a
        joint will never collide.
    allow_auto_disable: Allows a model to auto-disable, which is means
        the physics engine can skip updating the model when the model is
        at rest. This parameter is only used by models with no joints.
    include: Include resources from a URI. This can be used to nest
        models.
    model: A nested model element
    frame: A frame of reference to which a pose is relative.
    pose: A position(x,y,z) and orientation(roll, pitch yaw) with
        respect to the specified frame.
    link: A physical link with inertia, collision, and visual
        properties. A link must be a child of a model, and any number of
        links may exist in a model.
    joint: A joint connections two links with kinematic and dynamic
        properties.
    plugin: A plugin is a dynamically loaded chunk of code. It can exist
        as a child of world, model, and sensor.
    gripper:
    name: A unique name for the model. This name must not match another
        model in the world.
    """

    class Meta:
        name = "model"

    static: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    self_collide: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    allow_auto_disable: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    include: List["Model.Include"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    model: List["Model"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    frame: List["Model.Frame"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    pose: Optional["Model.Pose"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    link: List[Link] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    joint: List[Joint] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plugin: List["Model.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    gripper: List["Model.Gripper"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Include:
        """Include resources from a URI.

        This can be used to nest models.

        Parameters
        ----------
        uri: URI to a resource, such as a model
        pose: Override the pose of the included model. A position and
            orientation in the global coordinate frame for the model.
            Position(x,y,z) and rotation (roll, pitch yaw) in the global
            coordinate frame.
        name: Override the name of the included model.
        static: Override the static value of the included model.
        plugin: A plugin is a dynamically loaded chunk of code. It can
            exist as a child of world, model, and sensor.
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        pose: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        static: bool = field(
            default=False,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        plugin: List["Model.Include.Plugin"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Plugin:
            """A plugin is a dynamically loaded chunk of code.

            It can exist as a child of world, model, and sensor.

            Parameters
            ----------
            any_element: This is a special element that should not be
                specified in an SDFormat file. It automatically copies
                child elements into the SDFormat element so that a
                plugin can access the data.
            name: A unique name for the plugin, scoped to its parent.
            filename: Name of the shared library to load. If the
                filename is not a full path name, the file will be
                searched for in the configuration paths.
            """

            any_element: List[object] = field(
                default_factory=list,
                metadata={
                    "type": "Wildcard",
                    "namespace": "##any",
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )
            filename: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

    @dataclass
    class Frame:
        """
        A frame of reference to which a pose is relative.

        Parameters
        ----------
        pose: A position(x,y,z) and orientation(roll, pitch yaw) with
            respect to the specified frame.
        name: Name of the frame. This name must not match another frame
            defined inside the parent that this frame is attached to.
        """

        pose: Optional["Model.Frame.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            frame: Name of frame which the pose is defined relative to.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            frame: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

    @dataclass
    class Pose:
        """
        Parameters
        ----------
        value:
        frame: Name of frame which the pose is defined relative to.
        """

        value: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        frame: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            },
        )

    @dataclass
    class Plugin:
        """A plugin is a dynamically loaded chunk of code.

        It can exist as a child of world, model, and sensor.

        Parameters
        ----------
        any_element: This is a special element that should not be
            specified in an SDFormat file. It automatically copies child
            elements into the SDFormat element so that a plugin can
            access the data.
        name: A unique name for the plugin, scoped to its parent.
        filename: Name of the shared library to load. If the filename is
            not a full path name, the file will be searched for in the
            configuration paths.
        """

        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Gripper:
        grasp_check: Optional["Model.Gripper.GraspCheck"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        gripper_link: List[str] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
                "min_occurs": 1,
            },
        )
        palm_link: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class GraspCheck:
            detach_steps: int = field(
                default=40,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            attach_steps: int = field(
                default=20,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            min_contact_count: int = field(
                default=2,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )



from dataclasses import dataclass, field
from typing import Optional

__NAMESPACE__ = "sdformat/v1.0/scene.xsd"


@dataclass
class Scene:
    class Meta:
        name = "scene"

    ambient: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    background: Optional["Scene.Background"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    shadows: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    fog: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    grid: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )

    @dataclass
    class Background:
        sky: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        rgba: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )



"""<element name="sdf" required="1">
  <description>SDFormat base element that can include one model, actor, light, or worlds. A user of multiple worlds could run parallel instances of simulation, or offer selection of a world at runtime.</description>

  <attribute name="version" type="string" default="1.8" required="1">
    <description>Version number of the SDFormat specification.</description>
  </attribute>

  <include filename="world.sdf" required="*"/>
  <include filename="model.sdf" required="0"/>
  <include filename="actor.sdf" required="0"/>
  <include filename="light.sdf" required="0"/>

</element> <!-- End SDF -->"""

from typing import List, Union, Dict, Any, Tuple
import warnings

from .base import ElementBase
from .world import World
from .actor import Actor
from .model import Model
from .light import Light
from ..exceptions import ParseError
from .... import transform as tf


class Sdf(ElementBase):
    """SDFormat Root Element

    This element is a container for multiple simulation worlds (can be one) or
    for a single fragment of a world (Model, Actor, Light).

    Parameters
    ----------
    payload : Union[List[World], Model, Light, Actor]
        The element contained in this SDF. This can be one :class:`Model`, one :class`Actor`, one
        :class:`Light`, or a list of :class`Worlds`.
    version : str
        The SDFormat version.
    worlds : List[World]
        .. deprecated:: SDFormat v1.8
            Worlds, models, lights, and/or actors can no longer be combined. Use
            `payload` instead.
        The worlds contained in this SDF.
    models : List[Model]
        .. deprecated:: SDFormat v1.8
            Starting with SDFormat v1.8 only a single model is supported. Use the
            `model` kwarg instead.

        The models contained in this SDF.
    lights : List[Light]
        .. deprecated:: SDFormat v1.8
            Starting with SDFormat v1.8 only a single light is supported. Use the
            `light` kwarg instead.

        The lights contained in this SDF.
    actors : List[Actor]
        .. deprecated:: SDFormat v1.8
            Starting with SDFormat v1.8 only a single actor is supported. Use the
            `actor` kwarg instead.

        The actors contained in this SDF.

    Attributes
    ----------
    worlds : List[World]
        The worlds contained in the SDF file.
    model: Model
        The model contained in the SDF file.
    light: Light
        The light contained in the SDF file.
    actor: Actor
        The actor contained in the SDF file.
    version : str
        The SDFormat version.
    models : List[Model]
        .. deprecated:: SDFormat v1.8
            Use the `Sdf.model` instead.
    lights : List[Light]
        .. deprecated:: SDFormat v1.8
            Use the `Sdf.light` instead.
    actors : List[Actor]
        .. deprecated:: SDFormat v1.8
            Use the `Sdf.actor` instead.

    """

    def __init__(
        self,
        *,
        payload: Union[List[World], Model, Light, Actor] = None,
        version: str = "1.8",
        worlds: List[World] = None,
        models: List[Model] = None,
        lights: List[Light] = None,
        actors: List[Actor] = None,
    ) -> None:
        super().__init__(sdf_version=version)
        self.version = version
        self.worlds: List[World] = []
        self._actors: List[Actor] = []
        self._models: List[Model] = []
        self._lights: List[Light] = []

        if self.sdf_version == "1.8":
            if isinstance(payload, list) and all(
                [isinstance(x, World) for x in payload]
            ):
                self.worlds = payload
            elif isinstance(payload, Actor):
                self._actors.append(payload)
            elif isinstance(payload, Model):
                self._models.append(payload)
            elif isinstance(payload, Light):
                self._lights.append(payload)
            else:
                raise ParseError("Invalid `Sdf` element.")
        elif version in ["1.7", "1.6", "1.5", "1.4", "1.3"]:
            self.worlds.extend(worlds)
            self._models.extend(models)
            self._lights.extend(lights)
            self._actors.extend(actors)
        else:
            raise ParseError("`Sdf` does not exist prior to SDFormat v1.3.")

        for model in self._models:
            if model.pose.relative_to is None:
                model.pose.relative_to = "world"

    @property
    def actor(self) -> Actor:
        try:
            return self._actors[0]
        except IndexError:
            return None

    @property
    def model(self) -> Model:
        try:
            return self._models[0]
        except IndexError:
            return None

    @property
    def light(self) -> Light:
        try:
            return self._lights[0]
        except IndexError:
            return None

    # deprecated properties
    @property
    def actors(self) -> List[Actor]:
        warnings.warn(
            "`sdf.actors` is deprecated since SDFormat v1.8. Use `sdf.actor` instead.",
            DeprecationWarning,
        )
        return self._actors

    @property
    def models(self) -> List[Model]:
        warnings.warn(
            "`sdf.models` is deprecated since SDFormat v1.8. Use `sdf.models` instead.",
            DeprecationWarning,
        )
        return self._models

    @property
    def lights(self) -> List[Light]:
        warnings.warn(
            "`sdf.lights` is deprecated since SDFormat v1.8. Use `sdf.lights` instead.",
            DeprecationWarning,
        )
        return self._lights

    @classmethod
    def from_specific(cls, sdf: Any, *, version: str) -> "Sdf":
        """Create a generic Sdf object from a specific one.

        Parameters
        ----------
        sdf : Any
            The SDF object that should be turned into a generic object. It
            can come from any version of the specific SDFormat bindings
            provided by scikit-bot.
        version : str
            The version of the given SDF element.

        Returns
        -------
        generic_sdf : Sdf
            The generic equivalent of the provided specific Sdf object.
        """

        if version == "1.8":
            payload: Union[List[World], Model, Light, Actor]
            if sdf.world is not None and len(sdf.world) > 0:
                payload = [World.from_specific(x, version=version) for x in sdf.world]
            elif sdf.actor is not None:
                payload = Actor.from_specific(sdf.actor, version=version)
            elif sdf.model is not None:
                payload = Model.from_specific(sdf.model, version=version)
            elif sdf.light is not None:
                payload = Light.from_specific(sdf.light, version=version)
            else:
                raise ParseError("Can not convert `sdf` element without payload.")

            return Sdf(payload=payload, version=version)
        else:
            return Sdf(
                worlds=[World.from_specific(x, version=version) for x in sdf.world],
                actors=[Actor.from_specific(x, version=version) for x in sdf.actor],
                models=[Model.from_specific(x, version=version) for x in sdf.model],
                lights=[Light.from_specific(x, version=version) for x in sdf.light],
                version=version,
            )

    def declared_frames(
        self,
    ) -> Dict[str, Union[List[Dict[str, tf.Frame]], Dict[str, tf.Frame]]]:
        declared_frames = {
            "worlds": [x.declared_frames() for x in self.worlds],
            "models": [x.declared_frames() for x in self._models],
        }

        if self.model is not None:
            declared_frames["model"] = declared_frames["models"][0]

        return declared_frames

    def to_static_graph(
        self,
        declared_frames: Dict[str, List[tf.Frame]],
        *,
        seed: int = None,
        shape: Tuple[int] = (3,),
        axis: int = -1,
    ) -> Dict[str, List[tf.Frame]]:
        graphs = {
            "worlds": list(),
            "models": list(),
        }

        for idx, world in enumerate(self.worlds):
            try:
                static_frames = declared_frames["worlds"][idx]
            except IndexError:
                break

            root = world.to_static_graph(
                static_frames, seed=seed, shape=shape, axis=axis
            )
            graphs["worlds"].append(root)

        for idx, model in enumerate(self._models):
            try:
                static_frames = declared_frames["models"][idx]
            except IndexError:
                break

            if "world" not in static_frames:
                static_frames["world"] = tf.Frame(3, name="world")

            root = model.to_static_graph(
                static_frames, seed=seed, shape=shape, axis=axis
            )
            graphs["models"].append(root)

            model.pose.to_static_graph(
                static_frames, "__model__", shape=shape, axis=axis
            )

        return graphs

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, List[tf.Frame]],
        *,
        seed: int = None,
        shape: Tuple[int] = (3,),
        axis: int = -1,
        apply_state: bool = True,
    ) -> Dict[str, List[tf.Frame]]:
        graphs = {
            "worlds": list(),
            "models": list(),
        }
        scaffold = self.declared_frames()
        self.to_static_graph(scaffold, seed=seed, shape=shape, axis=axis)

        for idx, world in enumerate(self.worlds):
            try:
                frames = declared_frames["worlds"][idx]
            except IndexError:
                break

            root = world.to_dynamic_graph(frames, seed=seed, shape=shape, axis=axis)
            graphs["worlds"].append(root)

        for idx, model in enumerate(self._models):
            try:
                frames = declared_frames["models"][idx]
            except IndexError:
                break

            static_frames = scaffold["models"][idx]

            if "world" not in frames:
                frames["world"] = tf.Frame(3, name="world")

            root = model.to_dynamic_graph(
                frames, seed=seed, shape=shape, axis=axis, _scaffolding=static_frames
            )
            graphs["models"].append(root)

        return graphs



from math import tan
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from .affine import AffineLink
from ._utils import scalar_project


class PerspectiveProjection(AffineLink):
    """Perspective projection in N-D.

    This link projects a N dimensional frame onto an M dimensional frame. Using
    the parent's origin as the center of projection. In its most common use
    this corresponds to a central projection, e.g., the projection of
    coordinates in a 3D world frame down to a 2D camera frame.

    This link computes the projection using pairs of ``directions`` and
    ``amounts`` (both batches of vectors). To compute each coordinate of a
    vector in the projected space the vector is first scalar projected onto the
    amount (vector). This determines distance from the projection's center. Then
    the vector is scalar projected onto the direction (vector) and the
    result is scaled (anti-)proportional to the distance from the projection's
    center.

    Parameters
    ----------
    directions : ArrayLike
        A batch of (subspace-)vectors onto which points will be projected. The
        vectors run along ``axis`` and the subspace runs along
        ``subspace_axis``. All other dimensions are considered batch dimensions.
        Often this is a normal basis of the projection's subspace, e.g., the
        the x and y axes of a camera's image plane expressed in the parent
        frame.
    amounts : ArrayLike
        A batch of vectors indicating the direction along which to measure
        distance from the projection center. Its shape must match
        ``directions.shape``. Often all amount vectors are pairwise linearly
        dependent, e.g., they all point in the direction a camera is facing.
    axis : int
        The axis along which the projection is computed. It's length is equal to
        the number of dimensions in the parent frame.
    subspace_axis : int
        The axis along which different directions and amounts are stacked. It's
        length is equal to the number of dimensions in the child frame. Note
        that this axis _must_ be present, even if vectors are projected down to
        1D; in this case, the this axis has length 1.



    Methods
    -------
    transform(x)
        Expresses the vector x (assumed to be given in the parent's frame) in
        the child's frame.


    See Also
    --------
    :class:`skbot.transform.FrustumProjection`, :class:`skbot.ignition.FrustumProjection`

    Notes
    -----
    The length of a single direction vector rescales this axis. For example, if you have
    a camera with a certain number of pixels then the length of the direction vector would
    reflect this.

    The length of a single amount vector determines the scaling of distance. For example, if
    you have a camera with a certain focal lengths (fx, fy) then the length of the amount vector
    would reflect this.

    """

    def __init__(
        self,
        directions: ArrayLike,
        amounts: ArrayLike,
        *,
        axis: int = -1,
        subspace_axis: int = -2
    ) -> None:
        self.directions = np.asarray(directions)
        self.amounts = np.asarray(amounts)

        # make data axis the last axis (more efficient and easier to handle)
        # also make subspace axis the second last axis
        self.directions = np.moveaxis(self.directions, [subspace_axis, axis], [-2, -1])
        self.amounts = np.moveaxis(self.amounts, [subspace_axis, axis], [-2, -1])

        super().__init__(self.directions.shape[axis], self.directions.ndim, axis=axis)

    def transform(self, x: ArrayLike) -> np.ndarray:
        """Transform x (given in parent frame) into the child frame.

        Parameters
        ----------
        x : ArrayLike
            A batch of vectors expressed in the parent's frame. The parent frame runs
            along ``axis`` specified in the constructor.

        Returns
        -------
        y : ArrayLike
            A batch of vectors expressed in the child's frame. The child frame runs
            along ``axis`` specified in the constructor.

        Notes
        -----
        This function requires the batch dimensions of ``x``, ``amounts``, and
        ``directions`` to be broadcastable. To make an example assume a
        projection from N dimensions to M dimensions. In the trivial case
        (single vector, single projection) there are no batch dimensions; shapes
        are what you'd expect: ``x.shape=(N,)``, ``amounts.shape = (M, N)``,
        ``directions.shape=(M, N)``. In the case of a batch of vectors and a
        single projection, batch dimensions must be broadcastable:
        ``x.shape=(batch, N)``, ``amounts.shape = (1, M, N)``,
        ``directions.shape=(1, M, N)``. In the case of a single single vector
        and multiple projections the same rule applies: ``x.shape=(1, N)``,
        ``amounts.shape = (batch, M, N)``, ``directions.shape=(batch, M, N)``.
        Other combinations are - of course - possible, too.
        """
        x = np.asarray(x, dtype=np.float64)
        x = np.moveaxis(x, self._axis, -1)

        # make x broadcastable with amounts/directions
        x = np.expand_dims(x, -2)

        scaling = scalar_project(x, self.amounts, axis=-1)
        scaling /= np.linalg.norm(self.amounts, axis=-1)

        projected = scalar_project(x, self.directions, axis=-1)
        projected /= scaling * np.linalg.norm(self.directions, axis=-1)

        return projected

    def __inverse_transform__(self, x: ArrayLike) -> np.ndarray:
        """Transform x (given in the child frame) into the parent frame.

        Parameters
        ----------
        x : ArrayLike
            The vector expressed in the childs's frame

        Returns
        -------
        y : ArrayLike
            The vector expressed in the parents's frame

        """
        raise NotImplementedError("A projection isn't invertable in general.")



from dataclasses import dataclass, field
from typing import List, Optional

__NAMESPACE__ = "sdformat/v1.7/material.xsd"


@dataclass
class Material:
    """
    The material of the visual element.

    Parameters
    ----------
    script: Name of material from an installed script file. This will
        override the color element if the script exists.
    shader:
    render_order: Set render order for coplanar polygons. The higher
        value will be rendered on top of the other coplanar polygons
    lighting: If false, dynamic lighting will be disabled
    ambient: The ambient color of a material specified by set of four
        numbers representing red/green/blue, each in the range of [0,1].
    diffuse: The diffuse color of a material specified by set of four
        numbers representing red/green/blue/alpha, each in the range of
        [0,1].
    specular: The specular color of a material specified by set of four
        numbers representing red/green/blue/alpha, each in the range of
        [0,1].
    emissive: The emissive color of a material specified by set of four
        numbers representing red/green/blue, each in the range of [0,1].
    double_sided: If true, the mesh that this material is applied to
        will be rendered as double sided
    pbr: Physically Based Rendering (PBR) material. There are two PBR
        workflows: metal and specular. While both workflows and their
        parameters can be specified at the same time, typically only one
        of them will be used (depending on the underlying renderer
        capability). It is also recommended to use the same workflow for
        all materials in the world.
    """

    class Meta:
        name = "material"

    script: Optional["Material.Script"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    shader: Optional["Material.Shader"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    render_order: float = field(
        default=0.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    lighting: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    ambient: str = field(
        default="0 0 0 1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    diffuse: str = field(
        default="0 0 0 1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    specular: str = field(
        default="0 0 0 1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    emissive: str = field(
        default="0 0 0 1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    double_sided: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    pbr: Optional["Material.Pbr"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )

    @dataclass
    class Script:
        """Name of material from an installed script file.

        This will override the color element if the script exists.

        Parameters
        ----------
        uri: URI of the material script file
        name: Name of the script within the script file
        """

        uri: List[str] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
                "min_occurs": 1,
            },
        )
        name: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Shader:
        """
        Parameters
        ----------
        normal_map: filename of the normal map
        type: vertex, pixel, normal_map_object_space,
            normal_map_tangent_space
        """

        normal_map: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        type: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Pbr:
        """Physically Based Rendering (PBR) material.

        There are two PBR workflows: metal and specular. While both
        workflows and their parameters can be specified at the same
        time, typically only one of them will be used (depending on the
        underlying renderer capability). It is also recommended to use
        the same workflow for all materials in the world.

        Parameters
        ----------
        metal: PBR using the Metallic/Roughness workflow.
        specular: PBR using the Specular/Glossiness workflow.
        """

        metal: Optional["Material.Pbr.Metal"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        specular: Optional["Material.Pbr.Specular"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Metal:
            """
            PBR using the Metallic/Roughness workflow.

            Parameters
            ----------
            albedo_map: Filename of the diffuse/albedo map.
            roughness_map: Filename of the roughness map.
            roughness: Material roughness in the range of [0,1], where 0
                represents a smooth surface and 1 represents a rough
                surface. This is the inverse of a specular map in a PBR
                specular workflow.
            metalness_map: Filename of the metalness map.
            metalness: Material metalness in the range of [0,1], where 0
                represents non-metal and 1 represents raw metal
            environment_map: Filename of the environment / reflection
                map, typically in the form of a cubemap
            ambient_occlusion_map: Filename of the ambient occlusion
                map. The map defines the amount of ambient lighting on
                the surface.
            normal_map: Filename of the normal map. The normals can be
                in the object space or tangent space as specified in the
                'type' attribute
            emissive_map: Filename of the emissive map.
            light_map: Filename of the light map. The light map is a
                prebaked light texture that is applied over the albedo
                map
            """

            albedo_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            roughness_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            roughness: str = field(
                default="0.5",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            metalness_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            metalness: str = field(
                default="0.5",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            environment_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            ambient_occlusion_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            normal_map: Optional["Material.Pbr.Metal.NormalMap"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            emissive_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            light_map: Optional["Material.Pbr.Metal.LightMap"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class NormalMap:
                """
                Parameters
                ----------
                value:
                type: The space that the normals are in. Values are:
                    'object' or 'tangent'
                """

                value: str = field(
                    default="",
                    metadata={
                        "required": True,
                    },
                )
                type: str = field(
                    default="tangent",
                    metadata={
                        "type": "Attribute",
                    },
                )

            @dataclass
            class LightMap:
                """
                Parameters
                ----------
                value:
                uv_set: Index of the texture coordinate set to use.
                """

                value: str = field(
                    default="",
                    metadata={
                        "required": True,
                    },
                )
                uv_set: int = field(
                    default=0,
                    metadata={
                        "type": "Attribute",
                    },
                )

        @dataclass
        class Specular:
            """
            PBR using the Specular/Glossiness workflow.

            Parameters
            ----------
            albedo_map: Filename of the diffuse/albedo map.
            specular_map: Filename of the specular map.
            glossiness_map: Filename of the glossiness map.
            glossiness: Material glossiness in the range of [0-1], where
                0 represents a rough surface and 1 represents a smooth
                surface. This is the inverse of a roughness map in a PBR
                metal workflow.
            environment_map: Filename of the environment / reflection
                map, typically in the form of a cubemap
            ambient_occlusion_map: Filename of the ambient occlusion
                map. The map defines the amount of ambient lighting on
                the surface.
            normal_map: Filename of the normal map. The normals can be
                in the object space or tangent space as specified in the
                'type' attribute
            emissive_map: Filename of the emissive map.
            light_map: Filename of the light map. The light map is a
                prebaked light texture that is applied over the albedo
                map
            """

            albedo_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            specular_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            glossiness_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            glossiness: str = field(
                default="0",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            environment_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            ambient_occlusion_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            normal_map: Optional["Material.Pbr.Specular.NormalMap"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            emissive_map: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            light_map: Optional["Material.Pbr.Specular.LightMap"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class NormalMap:
                """
                Parameters
                ----------
                value:
                type: The space that the normals are in. Values are:
                    'object' or 'tangent'
                """

                value: str = field(
                    default="",
                    metadata={
                        "required": True,
                    },
                )
                type: str = field(
                    default="tangent",
                    metadata={
                        "type": "Attribute",
                    },
                )

            @dataclass
            class LightMap:
                """
                Parameters
                ----------
                value:
                uv_set: Index of the texture coordinate set to use.
                """

                value: str = field(
                    default="",
                    metadata={
                        "required": True,
                    },
                )
                uv_set: int = field(
                    default=0,
                    metadata={
                        "type": "Attribute",
                    },
                )



from dataclasses import dataclass, field
from typing import Optional

__NAMESPACE__ = "sdformat/v1.2/light.xsd"


@dataclass
class Light:
    """
    The light element describes a light source.

    Parameters
    ----------
    cast_shadows: When true, the light will cast shadows.
    pose: A position and orientation in the global coordinate frame for
        the light.
    diffuse: Diffuse light color
    specular: Specular light color
    attenuation: Light attenuation
    direction: Direction of the light, only applicable for spot and
        directional lights.
    spot: Spot light parameters
    name: A unique name for the light.
    type: The light type: point, directional, spot.
    """

    class Meta:
        name = "light"

    cast_shadows: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    pose: str = field(
        default="0 0 0 0 0 0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
        },
    )
    diffuse: str = field(
        default="1 1 1 1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    specular: str = field(
        default=".1 .1 .1 1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    attenuation: Optional["Light.Attenuation"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    direction: str = field(
        default="0 0 -1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
        },
    )
    spot: Optional["Light.Spot"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: str = field(
        default="__default__",
        metadata={
            "type": "Attribute",
        },
    )
    type: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Attenuation:
        """
        Light attenuation.

        Parameters
        ----------
        range: Range of the light
        linear: The linear attenuation factor: 1 means attenuate evenly
            over the distance.
        constant: The constant attenuation factor: 1.0 means never
            attenuate, 0.0 is complete attenutation.
        quadratic: The quadratic attenuation factor: adds a curvature to
            the attenuation.
        """

        range: float = field(
            default=10.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        linear: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        constant: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        quadratic: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Spot:
        """
        Spot light parameters.

        Parameters
        ----------
        inner_angle: Angle covered by the bright inner cone
        outer_angle: Angle covered by the outer cone
        falloff: The rate of falloff between the inner and outer cones.
            1.0 means a linear falloff, less means slower falloff,
            higher means faster falloff.
        """

        inner_angle: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        outer_angle: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        falloff: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )



from typing import List, Any, Tuple, Union

from .base import BoolElement, ElementBase, Pose, StringElement
from .plugin import Plugin
from ..exceptions import ParseError
from ... import fuel
from .. import sdformat


class Include(ElementBase):
    def __init__(
        self,
        *,
        uri: str,
        name: str = None,
        static: bool = None,
        pose: Pose = None,
        plugins: List["Plugin"] = None,
        placement_frame: str = None,
        sdf_version: str,
    ) -> None:
        super().__init__(sdf_version=sdf_version)
        self.uri = uri
        self.name = name
        self.static = static
        self.pose = pose
        self.plugins = [] if plugins is None else plugins
        self.placement_frame = placement_frame

        if self.placement_frame is not None:
            if self.pose is None:
                raise ParseError(
                    "`Include`s that specify `Include.placement_frame`"
                    " must also specify `Include.pose`."
                )

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "Include":
        # include_args = {
        #     "uri": specific.uri,
        #     "static": specific.static,
        #     "pose": Pose.from_specific(specific.pose, version=version),
        #     "plugins": [Plugin.from_specific(x, version=version) for x in specific.plugin],
        # }

        args_with_default = {
            "name": StringElement,
            "uri": StringElement,
            "static": BoolElement,
            "pose": Pose,
            "placement_frame": StringElement,
        }
        list_args = {"plugin": ("plugins", Plugin)}
        standard_args = cls._prepare_standard_args(
            specific, args_with_default, list_args, version=version
        )

        # if version not in ["1.0", "1.2", "1.3", "1.4", "1.5", "1.6", "1.7"]:
        #     include_args["placement_frame"] = specific.placement_frame

        return Include(
            **standard_args,
            sdf_version=version,
        )

    def resolve(self, *, priority: Tuple[str, str, str] = ("model", "actor", "light")):
        """Resolve the include.

        This function loads the included fragment as a generic :class:`Actor`,
        :class:`Model`, or :class:`Light`.

        Parameters
        ----------
        priority : str
            The order in which to check the included SDF for fragments. If
            multiple fragments are defined in the same SDF only the first
            fragment is returned. It is a 3-tuple with values `model`,
            `actor`, `light` in any order.

        Returns
        -------
        generic_sdf : Union["Model", "Light", "Actor"]
            The included fragment.

        """

        # avoid circular import while keeping resolve
        # as a method of include
        from .actor import Actor
        from .model import Model
        from .light import Light

        if not self.uri.startswith("https://fuel.ignitionrobotics.org"):
            raise ParseError("Includes from non-fuel sources are not implemented yet.")

        sdf_string = fuel.get_fuel_model(self.uri)
        specific_sdf = sdformat.loads(sdf_string)

        for el in priority:
            fragment = getattr(specific_sdf, el)

            if fragment is None:
                continue

            if isinstance(fragment, list):
                if len(fragment) == 0:
                    continue
                fragment = fragment[0]

            break
        else:
            raise ParseError(f"No fragments found at  `{self.uri}`.")

        if el == "model":
            model = Model.from_specific(fragment, version=specific_sdf.version)
            if self.name is not None:
                model.name = self.name
            if self.static is not None:
                model.static = self.static
            if self.pose is not None:
                model.pose = self.pose
                model._origin.pose = self.pose
            if self.placement_frame is not None:
                model.placement_frame = self.placement_frame
            model.plugins.extend(self.plugins)
            return model
        elif el == "actor":
            actor = Actor.from_specific(fragment, version=specific_sdf.version)
            raise NotImplementedError()
        else:
            light = Light.from_specific(fragment, version=specific_sdf.version)
            raise NotImplementedError()



import numpy as np
from scipy.interpolate import interp1d
from numpy.typing import ArrayLike
from typing import Optional


def linear_trajectory(
    t: ArrayLike,
    control_points: ArrayLike,
    *,
    t_control: Optional[ArrayLike] = None,
    t_min: float = 0,
    t_max: float = 1
) -> np.ndarray:
    """Evaluate the trajectory given by control_points at t using linear
    interpolation.

    ``linear_trajectory`` constructs a piece-wise linear trajectory using the
    given control points and then evaluates the resulting trajectory at ``t``.
    By default, control points are spaced out evenly in the interval ``[t_min,
    t_max]`` where ``t=t_min`` results in ``control_points[0]`` and ``t=t_max``
    results in ``control_poins[-1]``. Alternatively, the spacing of control
    points can be controlled manually by specifying ``t_control``, which
    implicitly specifies ``t_min`` and ``t_max``.

    Parameters
    ----------
    t : ArrayLike
        An array containing positions at which to evaluate the trajectory.
        Elements of ``t`` must be within ``[t_min, t_max]``.
    control_points : ArrayLike
        A batch of control points used to construct the trajectory. The first
        dimension of the array is interpreted as batch dimension and the
        remaining dimensions are used to interpolate between. By default,
        control points are equally spaced within ``[t_min, t_max]`` unless
        ``t_control`` is given explicitly.
    t_control : ArrayLike
        A sequence of strictly increasing floats determining the position of the
        control points along the trajectory. None by default, which results in
        an equidistant spacing of points.
    t_min : float
        Minimum value of the trajectories parametrization. Must be smaller than
        ``t_max``.If ``t_control`` is set, this value is ignored in favor of
        ``t_min=t_control[0]``.
    t_max : float
        Maximum value of the trajectories parametrization. Must be larger than
        ``t_min``. If ``t_control`` is set, this value is ignored in favor of
        ``t_max=t_control[-1]``.

    Returns
    -------
    position : ArrayLike
        The value of the trajectory at ``t``.

    Notes
    -----
    Repeated evaluation of single points on the trajectory, i.e. repeatedly
    calling this function with a scalar ``t``, is possible, but will repeatedly
    reconstruct the trajectory, which can lead to unnecessary slowdown. For
    better performance, it is preferred to use an array-like ``t``.

    Examples
    --------

    .. plot::
        :include-source:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from skbot.trajectory import linear_trajectory
        >>> t1 = np.linspace(0, 2*np.pi, 10)
        >>> control_points = np.stack((np.cos(t1), np.sin(t1)), axis=1)
        >>> t2 = np.linspace(0, 2*np.pi, 100)
        >>> trajectory = linear_trajectory(t2, control_points, t_min=0, t_max=2*np.pi)
        >>> fig, ax = plt.subplots()
        >>> ax.plot(trajectory[:,0], trajectory[:,1], control_points[:,0], control_points[:,1], 'o')
        >>> fig.legend(('Trajectory', 'Control Points'))
        >>> plt.show()


    """

    t = np.asarray(t)
    control_points = np.asarray(control_points)

    if t_control is None:
        t_control = np.linspace(t_min, t_max, len(control_points), dtype=np.float_)
    else:
        t_control = np.asarray(t_control)
        t_min = t_control[0]
        t_max = t_control[1]

    position = interp1d(t_control, control_points, axis=0)(t)

    return position



"""
Manage n-dimensional coordinate transformations via directed graphs.

``skbot.transform`` allows you to change the representation of a vector from one
coordinate system to another, i.e., it transforms the basis of the vector.
Contrary to similar libraries, this is not limited to 3D or affine
transformations, but can transform in N dimensions and between different
dimensions, too. Additionally, you can create chains - more precisely directed
graphs - of transformations between different frames, which allows you to
express quite complicated transformations.

If you come from a robotics background, this module is analogous to ROS tf2,
but works in (and between) n-dimensions. This means it naturally includes
projections, e.g. world space to pixel space, and it allows you to use more
esoteric transformations like spherical coordinates, too.


Examples
--------

>>> import skbot.transform as tf
>>> import numpy as np

Forward kinematics on a 1 DoF robot arm in 2D

>>> # setup of the arm
>>> link = tf.Translation((1, 0))
>>> joint = tf.Rotation((1, 0), (0, 1))
>>> tool_frame = tf.Frame(2)
>>> joint_frame = link(tool_frame)
>>> world = joint(joint_frame)
>>> joint.angle = 0
>>> # FK of initial position
>>> tool_pos = tool_frame.transform((0, 0), to_frame=world)
>>> assert np.allclose(tool_pos, (1, 0))
>>> # Set new joint angle and compute new FK
>>> joint.angle = np.pi / 2
>>> tool_pos = tool_frame.transform((0, 0), to_frame=world)
>>> assert np.allclose(tool_pos, (0, 1))

Base Elements
-------------

.. autosummary::
    :template: transform_class.rst
    :toctree:

    Frame
    Link

Available Transformations
-------------------------

.. rubric:: nD Links

.. autosummary::
    :template: transform_class.rst
    :toctree:

    AffineSpace
    Translation
    Rotation
    PerspectiveProjection

.. rubric:: 3D Links

.. autosummary::
    :template: transform_class.rst
    :toctree:

    EulerRotation
    QuaternionRotation
    RotvecRotation
    FrustumProjection
    RotationalJoint
    PrismaticJoint

.. rubric:: Other Links

.. autosummary::
    :template: transform_class.rst
    :toctree:

    CustomLink
    CompundLink
    InvertLink

"""

from .base import (
    Frame,
    Link,
    CustomLink,
    InvertLink,
    CompundLink,
)

from .functions import (
    # basic transformations
    translate,
    rotate,
    reflect,
    shear,
    scale,
)

from .affine import Translation, Rotation, Inverse, AffineSpace
from .projections import PerspectiveProjection
from .utils3d import (
    FrustumProjection,
    EulerRotation,
    QuaternionRotation,
    RotvecRotation,
)
from .joints import RotationalJoint, PrismaticJoint

from . import metrics

__all__ = [
    # Core Classes for Frame Management
    "Frame",
    "metrics",  # transform chain metrics
    "Link",
    # nD Links
    "AffineSpace",
    "Rotation",
    "Translation",
    "PerspectiveProjection",
    # 3D Links
    "EulerRotation",
    "QuaternionRotation",
    "RotationalJoint",
    "PrismaticJoint",
    "RotvecRotation",
    "FrustumProjection",
    # Other Links
    "CustomLink",
    "InvertLink",
    "CompundLink",
    "Inverse",
    # basic transformation functions
    "translate",
    "rotate",
    "reflect",
    "shear",
    "scale",
]



""" Interface with `ROS <https://www.ros.org/>`_.

Functions
---------

.. autosummary::
    :toctree: _autosummary

    create_frame_graph

"""

from .generic import create_frame_graph

__all__ = ["create_frame_graph"]



from .actor import Actor
from .collision import Collision
from .geometry import Geometry
from .joint import Joint
from .light import Light
from .link import Link
from .model import Model
from .physics import Physics
from .scene import Scene
from .sdf import Sdf
from .sensor import Sensor
from .state import State
from .visual import Visual
from .world import World

__all__ = [
    "Actor",
    "Collision",
    "Geometry",
    "Joint",
    "Light",
    "Link",
    "Model",
    "Physics",
    "Scene",
    "Sdf",
    "Sensor",
    "State",
    "Visual",
    "World",
]



from dataclasses import dataclass, field
from typing import List, Optional
from .joint import Joint
from .link import Link

__NAMESPACE__ = "sdformat/v1.2/actor.xsd"


@dataclass
class Actor:
    """
    Parameters
    ----------
    pose: Origin of the actor
    skin:
    animation:
    script:
    link: A physical link with inertia, collision, and visual
        properties. A link must be a child of a model, and any number of
        links may exist in a model.
    joint: A joint connections two links with kinematic and dynamic
        properties.
    plugin: A plugin is a dynamically loaded chunk of code. It can exist
        as a child of world, model, and sensor.
    name:
    static:
    """

    class Meta:
        name = "actor"

    pose: str = field(
        default="0 0 0 0 0 0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
        },
    )
    skin: Optional["Actor.Skin"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    animation: List["Actor.Animation"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
            "min_occurs": 1,
        },
    )
    script: Optional["Actor.Script"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    link: List[Link] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
            "min_occurs": 1,
        },
    )
    joint: List[Joint] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plugin: List["Actor.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    static: bool = field(
        default=False,
        metadata={
            "type": "Attribute",
        },
    )

    @dataclass
    class Skin:
        filename: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Animation:
        filename: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        interpolate_x: bool = field(
            default=False,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Script:
        loop: bool = field(
            default=True,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        delay_start: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        auto_start: bool = field(
            default=True,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        trajectory: List["Actor.Script.Trajectory"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Trajectory:
            waypoint: List["Actor.Script.Trajectory.Waypoint"] = field(
                default_factory=list,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            id: Optional[int] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )
            type: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class Waypoint:
                time: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                pose: str = field(
                    default="0 0 0 0 0 0",
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )

    @dataclass
    class Plugin:
        """A plugin is a dynamically loaded chunk of code.

        It can exist as a child of world, model, and sensor.

        Parameters
        ----------
        any_element: This is a special element that should not be
            specified in an SDFormat file. It automatically copies child
            elements into the SDFormat element so that a plugin can
            access the data.
        name: A unique name for the plugin, scoped to its parent.
        filename: Name of the shared library to load. If the filename is
            not a full path name, the file will be searched for in the
            configuration paths.
        """

        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )



from itertools import chain
from typing import List, Any, Dict, Tuple
import warnings
from itertools import chain


from .base import (
    BoolElement,
    ElementBase,
    NamedPoseBearing,
    PoseBearing,
    Pose,
    StringElement,
)
from .link import Link
from .include import Include
from .frame import Frame
from .joint import Joint
from .plugin import Plugin
from .gripper import Gripper
from .actor import Actor
from .light import Light
from ..exceptions import ParseError
from .... import transform as tf
from .origin import Origin


class Model(ElementBase):
    """A physical object in the simulation

    The model element defines a complete robot or any other physical object.
    This includes (but is not limited to) its appearance, collision,
    skeleton/kinematic chain.

    Parameters
    ----------
    name : str
        A unique name for the model. This name must not match another model in
        the world.
    canonical_link : str
        The name of the model's canonical link, to which the model's implicit
        coordinate frame is attached. If unset or set to an empty string,
        the first link element listed as a child of this model is chosen
        as the canonical link.

        .. versionadded:: SDFormat v1.7
    placement_frame : str
        The frame inside this model whose pose will be set by the pose element
        of the model. Defaults to  ``__model__`` (the implicit model frame).

        .. versionadded:: SDFormat v1.8
    static : bool
        If True (default), the model is immovable. Otherwise the model is
        simulated in the dynamics engine.
    self_collide : bool
        If True, all links in the model will collide with each other
        (except those connected by a joint). Can be overridden by the link or
        collision element self_collide property. Two links within a model will
        collide if link1.self_collide OR link2.self_collide. Links connected by
        a joint will never collide. The default is ``False``.

        .. versionadded:: SDFormat v1.5
    allow_auto_disable : bool
        Allows a model to auto-disable, which is means the physics engine can
        skip updating the model when the model is at rest. This parameter is
        only used by models with no joints.

        .. versionadded:: SDFormat v1.2
    frames : List[Frame]
        A list of frames of reference in which poses may be expressed.

        .. versionadded:: SDFormat v1.5
    pose : Pose
        The model's initial position (x,y,z) and orientation (roll, pitch, yaw).

        .. versionadded:: SDFormat v1.2
    links : List[Link]
        A list of links. Each link represents a rigid body with inertia,
        collision, and visual properties.

        .. versionchanged:: SDFormat v1.5
            A model may now have 0 links.
    joints : List[Joint]
        A list of joints. Each joint specifies a kinematic and/or dynamic
        constraints between two links.
    plugins : List[Plugin]
        A list of plugins used to customize the runtime behavior of the
        simulation.
    grippers: List[Gripper],
        A list of grippers.
    includes : List[Include]
        A list of references to other SDF files that contain :class:`Model`s to
        include as nested models.

        .. versionadded:: SDFormat v1.5
    models : List[Model]
        A list of models nested within this model.

        .. versionadded:: SDFormat v1.5
    enable_wind : bool
        If set to true, all links in the model will be affected by the wind. Can
        be overriden by the link wind property.

        .. versionadded:: SDFormat v1.6
    sdf_version : str
        The SDFormat version to use when constructing this element.
    origin : Origin
        The model's origin.

        .. deprecated:: SDFormat v1.2
            Use ``Model.pose`` instead.


    Attributes
    ----------
    name : str
        See ``Parameters`` section.
    canonical_link : str
        See ``Parameters`` section.
    placement_frame : str
        See ``Parameters`` section.
    static : bool
        See ``Parameters`` section.
    self_collide : bool
        See ``Parameters`` section.
    allow_auto_disable : bool
        See ``Parameters`` section.
    frames : List[Frame]
        See ``Parameters`` section.
    pose : Pose
        See ``Parameters`` section.
    links : List[Link]
        See ``Parameters`` section.
    joints : List[Joint]
        See ``Parameters`` section.
    plugins : List[Plugin]
        See ``Parameters`` section.
    grippers: List[Gripper],
        See ``Parameters`` section.
    includes : List[Include]
        See ``Parameters`` section.
    models : List[Model]
        See ``Parameters`` section.
    enable_wind : bool
        See ``Parameters`` section.

    """

    def __init__(
        self,
        *,
        name: str,
        canonical_link: str = None,
        placement_frame: str = "__model__",
        static: bool = False,
        self_collide: bool = False,
        allow_auto_disable: bool = True,
        frames: List[Frame] = None,
        pose: Pose = None,
        links: List[Link],
        joints: List[Joint],
        plugins: List[Plugin],
        grippers: List[Gripper],
        includes: List[Include] = None,
        models: List["Model"] = None,
        enable_wind: bool = False,
        sdf_version: str,
        origin: Origin = None,
    ) -> None:
        super().__init__(sdf_version=sdf_version)

        self.name = name

        if placement_frame == "":
            placement_frame = None
        if placement_frame is None:
            placement_frame = "__model__"
        self.placement_frame = placement_frame
        self.static = static
        self.self_collide = self_collide
        self.allow_auto_disable = allow_auto_disable
        self.frames = [] if frames is None else frames

        if origin is None:
            self._origin = Origin(sdf_version=sdf_version)
        elif sdf_version == "1.0":
            self._origin = origin
        else:
            warnings.warn("`origin` is deprecated. Use `Model.pose` instead.")
            self._origin = origin

        if len(links) == 0 and sdf_version in ["1.0", "1.2", "1.3", "1.4"]:
            raise ValueError("A `Model` must specify at least one `Link`.")
        else:
            self.links = links

        if sdf_version == "1.0":
            self.pose = self._origin.pose
        elif pose is None:
            self.pose = Pose(sdf_version=sdf_version)
        else:
            self.pose = pose
        self._origin.pose = self.pose

        self.joints = joints
        self.plugins = plugins
        self.grippers = grippers
        self.models = [] if models is None else models
        self.enable_wind = enable_wind

        for include in includes:
            fragment = include.resolve()
            if isinstance(fragment, Model):
                self.models.append(fragment)
            else:
                raise ValueError("`Model.include` can only be used to include models.")

        if canonical_link == "":
            canonical_link = None
        if self.static:
            self.canonical_link = "world"
        elif canonical_link is not None:
            self.canonical_link = canonical_link
        elif len(self.links) > 0:
            self.canonical_link = links[0].name
        elif len(self.models) > 0:
            model_name = self.models[0].name
            self.canonical_link = f"{model_name}"
        else:
            raise ValueError(
                "`Model` must specify `canonical_link` or have at least one `link`."
            )

        for el in chain(
            self.models,
            self.frames,
            self.links,
            self.joints
            # lights
        ):
            if el.pose.relative_to is None:
                el.pose.relative_to = "__model__"

        for frame in self.frames:
            if frame.attached_to is None:
                frame.attached_to = "__model__"

    @property
    def origin(self):
        warnings.warn(
            "`Model.origin` is deprecated since SDFormat v1.2. Use `Model.pose` instead."
        )
        return self._origin

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "Model":
        model_args = {"name": specific.name, "sdf_version": version}

        elements_with_default = {
            "canonical_link": StringElement,
            "placement_frame": StringElement,
            "static": BoolElement,
            "self_collide": BoolElement,
            "allow_auto_disable": BoolElement,
            "origin": Origin,
            "pose": Pose,
            "enable_wind": BoolElement,
        }
        list_elements = {
            "frame": ("frames", Frame),
            "link": ("links", Link),
            "joint": ("joints", Joint),
            "plugin": ("plugins", Plugin),
            "gripper": ("grippers", Gripper),
            "include": ("includes", Include),
            "model": ("models", Model),
        }

        standard_args = cls._prepare_standard_args(
            specific, elements_with_default, list_elements, version=version
        )
        model_args.update(standard_args)

        # if hasattr(specific, "placement_frame") and :
        #     model_args["placement_frame"]

        return Model(**model_args)

    def declared_frames(self) -> Dict[str, tf.Frame]:
        declared_frames = {"__model__": tf.Frame(3, name=self.name)}

        for el in chain(self.frames, self.links, self.joints):
            declared_frames.update(el.declared_frames())

        for model in self.models:
            model_frames = model.declared_frames()
            declared_frames[model.name] = model_frames["__model__"]
            for name, frame in model_frames.items():
                declared_frames[f"{model.name}::{name}"] = frame

        return declared_frames

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
    ) -> tf.Frame:
        for model in self.models:
            scope = {
                name.split("::", 1)[1]: frame
                for name, frame in declared_frames.items()
                if name.startswith(f"{model.name}::")
            }
            scope["world"] = declared_frames["world"]
            model.to_static_graph(scope, seed=seed, shape=shape, axis=axis)

        for model in self.models:
            link = model.pose.to_tf_link()
            parent_name = model.pose.relative_to
            child_name = f"{model.name}::{model.placement_frame}"

            parent = declared_frames[parent_name]
            child = declared_frames[child_name]
            link(child, parent)

        for el in chain(self.frames, self.links, self.joints):
            el.to_static_graph(declared_frames, seed=seed, shape=shape, axis=axis)
            el.pose.to_static_graph(declared_frames, el.name, shape=shape, axis=axis)

        return declared_frames["__model__"]

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple[int] = ...,
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame],
    ) -> tf.Frame:
        for model in self.models:
            scaffold_scope = {
                name.split("::", 1)[1]: frame
                for name, frame in _scaffolding.items()
                if name.startswith(f"{model.name}::")
            }
            scaffold_scope["world"] = _scaffolding["world"]

            scope = {
                name.split("::", 1)[1]: frame
                for name, frame in declared_frames.items()
                if name.startswith(f"{model.name}::")
            }
            scope["world"] = declared_frames["world"]
            model.to_dynamic_graph(
                scope,
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=scaffold_scope,
            )

        if self.static:
            parent = declared_frames["world"]
            parent_static = _scaffolding["world"]
        else:
            parent = declared_frames[self.canonical_link]
            parent_static = _scaffolding[self.canonical_link]
        child = declared_frames["__model__"]
        child_static = _scaffolding["__model__"]
        tf.CompundLink(parent_static.links_between(child_static))(parent, child)

        for el in chain(self.frames, self.links, self.joints):
            el.to_dynamic_graph(
                declared_frames,
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=_scaffolding,
            )

        return declared_frames["__model__"]



import setuptools

install_requires = ["numpy == 1.21.2", "scipy == 1.7.1"]

extras_require = {
    "ignition": [
        "pyzmq == 22.3.0",
        "betterproto == 1.2.5",
        "psutil == 5.8.0",
        "requests == 2.26.0",
        "xsdata == 21.9",
        "lxml == 4.6.3",
        "cachetools == 4.2.4",
    ],
    "docs": [
        "sphinx == 4.2.0",
        "numpydoc == 1.1.0",
        "sphinx-autodoc-typehints == 1.12.0",
        "matplotlib == 3.4.3",
        "pydata-sphinx-theme == 0.7.1",
        "cachetools == 4.2.4",
        "sphinx-gallery == 0.10.0",
        "xsdata == 21.9",
        "lxml == 4.6.3",
    ],
    "linting": ["flake8 == 3.9.2", "black == 21.9b0"],
    "testing": ["pytest == 6.2.5", "coverage[toml] == 6.0.2"],
    "dev": [
        "lxml-stubs == 0.2.0",
        "python-semantic-release == 7.19.2",
    ],
    "build": [
        "build == 0.7.0",
        "twine == 3.4.2",
    ],
}

# Note: This is a shim, because I am used to developing
# under pip install -e . which PEP517 doesn't support (yet?)
if __name__ == "__main__":
    setuptools.setup(
        install_requires=install_requires,
        extras_require=extras_require,
        package_data={"skbot": ["ignition/sdformat/schema/**/*.xsd"]},
    )



from dataclasses import dataclass, field
from typing import Optional

__NAMESPACE__ = "sdformat/v1.6/scene.xsd"


@dataclass
class Scene:
    """
    Specifies the look of the environment.

    Parameters
    ----------
    ambient: Color of the ambient light.
    background: Color of the background.
    sky: Properties for the sky
    shadows: Enable/disable shadows
    fog: Controls fog
    grid: Enable/disable the grid
    origin_visual: Show/hide world origin indicator
    """

    class Meta:
        name = "scene"

    ambient: str = field(
        default="0.4 0.4 0.4 1.0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    background: str = field(
        default=".7 .7 .7 1",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
        },
    )
    sky: Optional["Scene.Sky"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    shadows: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    fog: Optional["Scene.Fog"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    grid: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    origin_visual: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )

    @dataclass
    class Sky:
        """
        Properties for the sky.

        Parameters
        ----------
        time: Time of day [0..24]
        sunrise: Sunrise time [0..24]
        sunset: Sunset time [0..24]
        clouds: Sunset time [0..24]
        """

        time: float = field(
            default=10.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        sunrise: float = field(
            default=6.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        sunset: float = field(
            default=20.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        clouds: Optional["Scene.Sky.Clouds"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Clouds:
            """
            Sunset time [0..24]

            Parameters
            ----------
            speed: Speed of the clouds
            direction: Direction of the cloud movement
            humidity: Density of clouds
            mean_size: Average size of the clouds
            ambient: Ambient cloud color
            """

            speed: float = field(
                default=0.6,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            direction: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            humidity: float = field(
                default=0.5,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            mean_size: float = field(
                default=0.5,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            ambient: str = field(
                default=".8 .8 .8 1",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                    "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
                },
            )

    @dataclass
    class Fog:
        """
        Controls fog.

        Parameters
        ----------
        color: Fog color
        type: Fog type: constant, linear, quadratic
        start: Distance to start of fog
        end: Distance to end of fog
        density: Density of fog
        """

        color: str = field(
            default="1 1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )
        type: str = field(
            default="none",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        start: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        end: float = field(
            default=100.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        density: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )



from typing import Union, List, Tuple, Dict
from itertools import chain

from ... import transform as tf
from .load_as_generic import loads_generic
from .generic_sdf.world import World


def to_frame_graph(
    sdf: str,
    *,
    unwrap: bool = True,
    insert_world_frame: bool = True,
    shape: Tuple[int] = (3,),
    axis: int = -1
) -> Union[tf.Frame, List[tf.Frame]]:
    """Create a frame graph from a sdformat string.

    .. versionadded:: 0.8.0
        Added the ability to limit loading to worlds
    .. versionadded:: 0.6.0
        This function has been added to the library.

    Parameters
    ----------
    sdf : str
        A SDFormat XML string describing one (or many) worlds.
    unwrap : bool
        If True (default) and the sdf only contains a single light, model, or
        world element return that element's frame. If the sdf contains multiple
        lights, models, or worlds a list of root frames is returned. If False,
        always return a list of frames.
    insert_world_frame : bool
        If ``False``, creation of frame graphs is restricted to world elements
        contained in the provided SDF. This is because non-world elements
        (simulation fragments) may refer to a ``world`` frame that is defined
        outside of the provided SDF and hence the full graph can't be
        determined. As a consequence, any ``model``, ``actor``, or ``light``
        elements are ignored.

        If ``True`` (default), this function will insert a ``world`` frame into
        the graph of each simulation fragment (non-world element) to allow
        smooth construction of the frame graph.

        The default is ``True``; however, it will change to ``False`` starting
        with scikit-bot v1.0.

    shape : tuple
        A tuple describing the shape of elements that the resulting graph should
        transform. This can be used to add batch dimensions to the graph, for
        example to perform vectorized computation on multiple instances of the
        same world in different states, or for batched coordinate
        transformation. Defaults to (3,), which is the shape of a single 3D
        vector in euclidian space.
    axis : int
        The axis along which elements are stored. The axis must have length 3
        (since SDFormat describes 3 dimensional worlds), and all other axis are
        considered batch dimensions. Defaults to -1.

    Returns
    -------
    frame_graph : Union[Frame, List[Frame]]
        A :class:`skbot.transform.Frame` or list of Frames depending on the value of
        ``unwrap`` and the number of elements in the SDF's root element.

    See Also
    --------
    :mod:`skbot.transform`

    Notes
    -----
    Frames inside the graph are named after the frames defined by the SDF. You can
    retrieve them by searching for them using :func:`skbot.transform.Frame.find_frame`.

    Joins are implicit within the frame graph. The joint frame that is attached
    to the child frame is named after the joint (<joint_name>), and the
    (implicit) joint frame attached to the parent is named
    (<joint_name>_parent). The link between the two frames can be retrieved via
    :func:`skbot.transform.Frame.links_between`. For example if there is a
    joint named "robot_joint0" its link can be retrieved using::

        child_frame = frame_graph.find_frame(".../robot_joint0")
        parent_frame = frame_graph.find_frame(".../robot_joint0_parent")
        link = child_frame.links_between(child_frame)[0]
    """

    root = loads_generic(sdf)
    declared_frames = root.declared_frames()
    dynamic_graphs = root.to_dynamic_graph(declared_frames, shape=shape, axis=axis)

    if insert_world_frame:
        candidates = dynamic_graphs.values()
    else:
        candidates = [dynamic_graphs["worlds"]]

    graphs = list()
    for x in chain([x for x in candidates]):
        graphs.extend(x)

    if len(graphs) == 0:
        raise ValueError("No graphs could be loaded from the provided SDF.")

    if unwrap and len(graphs) == 1:
        return graphs[0]
    else:
        return graphs



from typing import Dict, List, Tuple
from xml.etree import ElementTree
import numpy as np
from dataclasses import dataclass
from typing import List, Dict
import requests
from urllib.parse import urlparse
from pathlib import Path
import warnings

from ... import transform as rtf


def _xpath_from_elements(elements: List[ElementTree.Element]) -> str:
    return ("/" if len(elements) > 0 else "") + "/".join(
        el.attrib["name"] if "name" in el.attrib else el.tag for el in elements
    )


def _fetch_include_uri(
    include_element: List[ElementTree.Element], *, fuel_download_path: str = None
) -> str:
    """Convert an include element into the full sdf.

    Includes Fuel support.

    Parameters
    ----------
    fuel_download_path : str
        If specified, download the full model from the fuel database into the
        specified directory. If None, only fetch the relevant model.sdf
        (in-memory).

    """
    uri = include_element.find("uri").text
    uri_parts = urlparse(uri)

    # include fuel model
    if uri_parts.scheme == "https" and uri_parts.netloc == "fuel.ignitionrobotics.org":
        # I reverse engineered this from the ign_fuel_tools C++ project
        # and from https://app.ignitionrobotics.org/api If you know a
        # cleaner way that doesn't involve adding a C++ to the codebase
        # just to download files please open an issue :)

        file_list: requests.Response = requests.get(uri + "/files")
        if file_list.status_code != 200:
            # Note: I would like to discover the latest version if
            # it isn't specified explicitly, but didn't manage to
            # work this out yet
            uri = uri + "/1"
            file_list: requests.Response = requests.get(uri + "/files")

        if file_list.status_code != 200:
            raise IOError(f"Could not download element from: {uri}")

        for item in file_list.json()["file_tree"]:
            if item["name"] == "model.sdf":
                location = uri + "/files" + item["path"]
                sdf_string = requests.get(location).content.decode("utf-8")
                break
        else:
            raise IOError(f"Could not get sdf from: {uri}")

        if fuel_download_path is not None:
            # tree all files
            full_file_list = list()
            file_queue = file_list.json()["file_tree"]
            while len(file_queue) > 0:
                element = file_queue.pop(0)
                if "children" in element.keys():
                    file_queue += element["children"]
                else:
                    full_file_list.append(element["path"])

            # TODO: figure out the default naming scheme for fuel downloads
            # to match the official fuel cache.
            base_dir = fuel_download_path / file_list.json()["name"]
            for file in full_file_list:
                location: Path = base_dir / file[1:]
                location.parent.mkdir(exist_ok=True, parents=True)
                file_request = requests.get(uri + "/files" + file)

                with open(location, "wb") as file_on_disk:
                    file_on_disk.write(file_request.content)

    elif uri_parts.scheme == "" or uri_parts.scheme == "file":
        with open(uri, "r") as sdf_file:
            sdf_string = sdf_file.read()

    else:
        raise IOError(f"Could not get sdf from: {uri}")

    return sdf_string


def create_frame_graph(sdf: str) -> Tuple[Dict[str, rtf.Frame], Dict[str, rtf.Link]]:
    """Create a frame graph from a sdformat string.

    .. deprecated:: 0.6.0
        This function will be removed in skbot v1.0. Use
        :func:``skbot.ignition.to_frame_graph`` instead. To find frames in the
        frame graph returned by ``to_frame_graph`` use
        :func:``skbot.transform.Frame.find_frame``. Links/Joints can be obtained
        from :func:``skbot.transform.Frame.links_between``.

    Parameters
    ----------
    sdformat: TextIO
        A text buffer containing the SDFormat XML.

    Returns
    -------
    frames : Dict[str, Frames]
        A dict of named frames. Keys and frame names correspond to each
        element's xpath. An xpath is a string representing the position of the
        element in the tree it is constructed from concatenating element tags
        using ``/`` except for elements that have a ``name`` attribute. In this
        case, the respective name is used instead of the tag. Examples:
        `/sdf/spot_light/pose` or `/sdf/world_1/my_robot/joint_1`

    links : Dict[str, Frames]
        A dict of (named) links in the graph.

    See Also
    --------
    :mod:`skbot.transform`

    """

    warnings.warn(
        """create_frame_graph is deprecated and will be removed in skbot v1.0 in
        favor of to_frame_graph. Check the docs for detailed migration
        instructions.
        """,
        DeprecationWarning,
    )

    root = ElementTree.fromstring(sdf)

    @dataclass
    class SdfQueueItem:
        element: ElementTree.Element
        parents: List[ElementTree.Element]

    @dataclass
    class PoseQueueItem:
        pose: np.ndarray
        parents: List[ElementTree.Element]
        relative_to: str = ""

    xpaths = dict()
    links = dict()
    pose_list: List[PoseQueueItem] = list()

    queue = [SdfQueueItem(root, list())]
    while len(queue) > 0:
        item: SdfQueueItem = queue.pop(0)

        for child in item.element:
            queue.append(SdfQueueItem(child, item.parents + [item.element]))

        # each elements gets an accompanying frame based on its xPath
        # if the element has a name, the frame will get that name, too
        name = item.element.attrib["name"] if "name" in item.element.attrib else None
        xpath_str = item.element.tag if name is None else name
        xpath = _xpath_from_elements(item.parents) + "/" + xpath_str

        if xpath in xpaths.keys():
            Warning(
                "non-unique xpaths found. Inspect the sdf file for errors and tread carefully."
            )

        xpaths[xpath] = rtf.Frame(3, name=xpath)

        # element specific graph modifications
        if item.element.tag == "pose":
            pose_tag = item.element
            pose = np.array(pose_tag.text.split(" "), dtype=np.float_)

            parent = _xpath_from_elements(item.parents[:-1])
            child = _xpath_from_elements(item.parents)
            pose_item = PoseQueueItem(pose, item.parents)
            if "relative_to" in pose_tag.keys():
                pose_item.relative_to = pose_tag.attrib["relative_to"]

            # we construct links at the end, because the "relative_to"
            # frame may not have been defined yet.
            pose_list.append(pose_item)

        elif item.element.tag == "include":
            # included sdf contains _exactly_ one child
            include_sdf_string = _fetch_include_uri(item.element)
            sdf_element = ElementTree.fromstring(include_sdf_string)[0]

            name_element = item.element.find("name")
            if name_element is not None:
                sdf_element.set("name", name_element.text)

            static_element = item.element.find("static")
            if name_element is not None:
                sdf_element.find("static").text = static_element.text

            # TODO: add support for pose and reference_frame

            queue.append(SdfQueueItem(sdf_element, item.parents))

    # all frames exist, add links
    # add pose-based (static) links to graph
    for item in pose_list:
        parent = _xpath_from_elements(item.parents[:-1])
        child = _xpath_from_elements(item.parents)

        if item.relative_to:
            # resolve relative parent including grandparents
            for idx in range(len(item.parents), 1, -1):
                parent_xpath = _xpath_from_elements(item.parents[:idx])
                xpath = parent_xpath + "/" + item.relative_to
                if xpath in xpaths:
                    parent = xpath
                    break
            else:
                raise RuntimeError(
                    f"The Frame '{item.parent}' does not exist among the parents of {item.child}."
                )

        parent_frame = xpaths[parent]
        child_frame = xpaths[child]
        if np.any(pose[3:]) != 0:
            rotated = rtf.EulerRotation("XYZ", pose[3:])(child_frame)
            rtf.Translation(pose[:3])(rotated, parent_frame)
        else:
            rtf.Translation(pose[:3])(child_frame, parent_frame)

    return xpaths, links



from dataclasses import dataclass, field
from typing import Optional
from .geometry import Geometry

__NAMESPACE__ = "sdformat/v1.0/collision.xsd"


@dataclass
class Collision:
    class Meta:
        name = "collision"

    max_contacts: int = field(
        default=10,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    mass: float = field(
        default=0.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    origin: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    geometry: Optional[Geometry] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    surface: Optional["Collision.Surface"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    laser_retro: float = field(
        default=0.0,
        metadata={
            "type": "Attribute",
        },
    )

    @dataclass
    class Surface:
        bounce: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        friction: Optional["Collision.Surface.Friction"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        contact: Optional["Collision.Surface.Contact"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Friction:
            ode: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

        @dataclass
        class Contact:
            ode: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )



"""
Interface with Ignitionrobotics_ libraries.


.. note::

    To use :mod:`skbot.ignition`, you have to install the latest version of
    Ignition_. In addition, you have to install
    additional dependencies via

    .. code-block:: bash

        pip install -e .[ignition]

:mod:`skbot.ignition` is a collection of functions to ease interfacing with the
Ignition software stack. Ignition is a collection of libraries that, when
combined, produce a general purpose simulator for robotic applications called
Ignition Gazebo. Ignition Gazebo superseeds Gazebo and, among many other
improvements, it is written as a simulation library instead of being a
stand-alone application. This allows for dynamic control of all aspects of the
simulation as well as full introspection.

Note that this module does not contain python bindings for ignition (and it does
not aim to do so). If you are looking for bindings check out
ScenarIO_,
which is a part of gym-ignition_.


.. currentmodule:: skbot.ignition

Functions
---------

.. rubric:: General

.. autosummary::
    :toctree:

    skbot.ignition.create_frame_graph
    skbot.ignition.download_fuel_model
    skbot.ignition.FrustumProjection
    skbot.ignition.get_fuel_model
    skbot.ignition.get_fuel_model_info
    skbot.ignition.Subscriber

.. rubric:: SDFormat Specific

.. note::
    Scikit-bot will automatically select the correct version when using
    :func:`skbot.ignition.sdformat.loads`; however, it is imporant that you are
    mindful of the version you are using, since SDFormat doesn't use SemVer and
    availability of elements may differ slightly across versions.

.. autosummary::
    :toctree:

    skbot.ignition.sdformat.dumps
    skbot.ignition.sdformat.get_version
    skbot.ignition.sdformat.loads
    skbot.ignition.sdformat.loads_generic
    skbot.ignition.sdformat.to_frame_graph


SDFormat XML
------------

.. note::
    You can find more documentation aboud SDFormat in the `official spec <http://sdformat.org/spec>`_.

.. rubric:: _`XSD Schema Files`

.. warning::
    The XSD bindings are (slightly) oppinionated and promote some optional
    elements to required elements with a default value. Check the content of
    this section for details.

Scikit-bot comes with XSD1.1 schema files that describe SDFormat. You can find their
latest versions inside the skbot/ignition/sdformat folder. You can view them on
GitHub and a copy can be found in your local version, too. Contrary to the
official SDFormat, schemas are provided for *all* SDF versions.

The XSD bindings are (slightly) oppinionated. If a SDF element is optional (the
spec denotes this as required=0 or required=*) but specifies a default value,
the schema promotes this to a *required* element with a default value. This is
done to allow the python bindings to provide you with auto-populated defaults
where possible.

.. rubric:: _`SDF Bindings`

Scikit-bot features a DOM-style parser and serializer for SDFormat XML. You can load
all SDFormat versions and scikit-bot will construct a class tree out of it.
Similarily, you can construct objects that represent SDF elements and scikit-bot can
serialize them into SDF. 

The API used here mimics the API used by the familiar `JSON parser <json>`_ or
`YAML parser <yaml>`_. The main difference is that this module returns an object
tree of dataclass objects, whilst JSON and YAML return dictionaries.

While the parser is imported together with :mod:`skbot.ignition`, the individual
models are imported on demand. This is done to keep import times low. To use the
bindings explicitly you must import them explicitly. Check the individual
bindings for documentation on how to do this:

.. autosummary::
    :template: sdformat_model.rst
    :toctree:
    :recursive:

    skbot.ignition.sdformat.bindings.v18
    skbot.ignition.sdformat.bindings.v17
    skbot.ignition.sdformat.bindings.v16
    skbot.ignition.sdformat.bindings.v15
    skbot.ignition.sdformat.bindings.v14
    skbot.ignition.sdformat.bindings.v13
    skbot.ignition.sdformat.bindings.v12
    skbot.ignition.sdformat.bindings.v10

Ignition Messages
-----------------

Scikit-bot provides python bindings to all Ignition messages. Messages are build
from the protocol buffer templates found in Ign-Msgs_. This allows you to decode
messages sent by the subscriber.

To build the bindings yourself, run

.. code-block:: bash

    git clone https://github.com/ignitionrobotics/ign-msgs.git
    mkdir ign-bindings
    protoc -I ign-msgs/proto/ --python_betterproto_out=ign-bindings ign-msgs/proto/ignition/msgs/*

This will produce a file called `ignition/msgs.py` in the chosen output location
which can be used to decode messages.

Members of ``skbot.ignition.messages`` are not documented here (the class is
fully auto-generated). Instead, an overview can be found in the `Ignition
documentation`_.

.. _gym-ignition: https://github.com/robotology/gym-ignition
.. _ScenarIO: https://robotology.github.io/gym-ignition/master/motivations/why_gym_ignition.html
.. _Ignition: https://ignitionrobotics.org/home
.. _`Ign-Msgs`: https://github.com/ignitionrobotics/ign-msgs
.. _`Ignition documentation`: https://ignitionrobotics.org/api/msgs/6.4/index.html
.. _Ignitionrobotics: https://ignitionrobotics.org/
.. _yaml: https://pyyaml.org/wiki/PyYAMLDocumentation
.. _json: https://docs.python.org/3/library/json.html
"""

from . import messages
from .subscriber import Subscriber
from .transformations import FrustumProjection
from .sdformat.create_frame_graph import create_frame_graph
from . import sdformat
from .fuel import get_fuel_model_info, download_fuel_model, get_fuel_model

__all__ = [
    "messages",
    "Subscriber",
    "FrustumProjection",
    "create_frame_graph",
    "sdformat",
    "get_fuel_model_info",
    "download_fuel_model",
    "get_fuel_model",
    "transform_graph_from_sdf",
]



import warnings
from itertools import chain
from typing import List, Any, Dict, Tuple


from .base import BoolElement, ElementBase, Pose
from .light import Light
from .frame import Frame
from .sensor import Sensor
from .origin import Origin
from .inertial import Inertial
from .collision import Collision
from .visual import Visual
from .sensor import Sensor
from .projector import Projector
from .audio_sink import AudioSink
from .audio_source import AudioSource
from .battery import Battery
from .particle_emitter import ParticleEmitter
from .... import transform as tf


class Link(ElementBase):
    """A rigid body.

    Parameters
    ----------
    name : str
        The name of the link. It must be unique within its scope (the containing
        model).
    gravity : bool
        If true (default), the link is affected by gravity.
    enable_wind : bool
        If true, the link is affected by the wind. Default: ``False``.

        .. versionadded:: SDFormat v1.6
    self_collide : bool
        If true, the link can collide with other links in the model. Any other
        link within a model will collide with this link if either
        ``self.self_collide == True`` or ``other.self_collide == True``.
        Further, links connected by a joint will never collide. Default:
        ``False``.
    kinematic : bool
        If True, the link is kinematic only. Default: `` False``
    pose : Pose
        The links's initial position (x,y,z) and orientation (roll, pitch, yaw).

        .. versionadded:: SDFormat 1.2
    must_be_base_link : bool
        If true, the link will have 6DOF and be a direct child of world.

        .. versionadded:: SDFormat 1.4
    velocity_decay : Link.VelocityDecay
        Exponential damping of the link's velocity.

        .. versionadded:: SDFormat 1.2
    inertial : Inertial
        The inertial properties of the link.
    collisions : List[Collision]
        The list of collision properties of a link. Note that this can be
        different from the visual properties of a link, for example, simpler
        collision models are often used to reduce computation time.
    visuals : List[Visual]
        A list of visual properties of the link. This element specifies the
        shape of the object (box, cylinder, etc.) for visualization purposes.
    sensors : List[Sensor]
        A list of sensors attached to this link.
    projectors : List[Projector]
        A list of projectors attached to this link.
    audio_sinks : List[AudioSink]
        A list of audio sinks attached to this link.

        .. versionadded:: SDFormat v1.4
    audio_sources : List[AudioSource]
        A list of audio sources attached to this link.

        .. versionadded:: SDFormat v1.4
    batteries : List[Battery]
        A list of batteries attached to this link.

        .. versionadded:: SDFormat v1.5
    lights : List[Light]
        A list of light sources attached to this link.

        .. versionadded:: SDFormat v1.6
    particle_emitters : List[ParticleEmitter]
        A list of particle emitters attached to this link. A particle emitter
        that can be used to describe fog, smoke, and dust.

        .. versionadded:: SDFormat v1.6
    sdf_version : str
        The SDFormat version to use when constructing this element.
    origin : Origin
        The link's origin.

        .. deprecated:: SDFormat v1.2
            Use `Link.pose` instead.
    damping : Link.Damping
        Exponential damping of the link's velocity.

        .. deprecated:: SDFormat 1.2
            Use `Link.velocity_decay` instead.
    frames : List["Frame"]
        A list of frames of reference in which poses may be expressed.

        .. deprecated:: SDFormat v1.7
            Use :attr:`Model.frame` instead.
        .. versionadded:: SDFormat v1.5


    Attributes
    ----------
    name : str
        See ``Parameters`` section.
    gravity : bool
        See ``Parameters`` section.
    enable_wind : bool
        See ``Parameters`` section.
    self_collide : bool
        See ``Parameters`` section.
    kinematic : bool
        See ``Parameters`` section.
    pose : Pose
        See ``Parameters`` section.
    must_be_base_link : bool
        See ``Parameters`` section.
    velocity_decay : Link.VelocityDecay
        See ``Parameters`` section.
    inertial : Inertial
        See ``Parameters`` section.
    collisions : List[Collision]
        See ``Parameters`` section.
    visuals : List[Visual]
        See ``Parameters`` section.
    sensors : List[Sensor]
        See ``Parameters`` section.
    projectors : List[Projector]
        See ``Parameters`` section.
    audio_sinks : List[AudioSink]
        See ``Parameters`` section.
    audio_sources : List[AudioSource]
        See ``Parameters`` section.
    batteries : List[Battery]
        See ``Parameters`` section.
    lights : List[Light]
        See ``Parameters`` section.
    particle_emitters : List[ParticleEmitter]
        See ``Parameters`` section.

    """

    def __init__(
        self,
        *,
        name: str,
        gravity: bool = True,
        enable_wind: bool = False,
        self_collide: bool = False,
        kinematic: bool = False,
        origin: Origin = None,
        pose: Pose = None,
        must_be_base_link: bool = False,
        damping: "Link.Damping" = None,
        velocity_decay: "Link.VelocityDecay" = None,
        frames: List["Frame"] = None,
        inertial: Inertial = None,
        collisions: List[Collision] = None,
        visuals: List[Visual] = None,
        sensors: List[Sensor] = None,
        projector: Projector = None,
        audio_sinks: List[AudioSink] = None,
        audio_sources: List[AudioSource] = None,
        batteries: List[Battery] = None,
        lights: List["Light"] = None,
        particle_emitters: List[ParticleEmitter] = None,
        sdf_version: str,
    ) -> None:
        super().__init__(sdf_version=sdf_version)
        self.name = name
        self.gravity = gravity
        self.enable_wind = enable_wind
        self.self_collide = self_collide
        self.kinematic = kinematic
        if origin is None:
            self._origin = Origin(sdf_version=sdf_version)
        elif sdf_version == "1.0":
            self._origin = origin
        else:
            warnings.warn("`origin` is deprecated. Use `pose` instead.")
            self._origin = origin
        if sdf_version == "1.0":
            self.pose = self._origin.pose
        elif pose is None:
            self.pose = Pose(sdf_version=sdf_version)
        else:
            self.pose = pose
        self.must_be_base_link = must_be_base_link
        if damping is None:
            self._damping = Link.Damping(sdf_version=sdf_version)
        elif sdf_version == "1.0":
            self._damping = damping
        else:
            warnings.warn("`damping` is deprecated. Use `velocity_decay` instead.")
            self._damping = damping
        if sdf_version == "1.0":
            self.velocity_decay = self._damping
        elif velocity_decay is None:
            self.velocity_decay = Link.VelocityDecay(sdf_version=sdf_version)
        else:
            self.velocity_decay = velocity_decay
        self._frames = [] if frames is None else frames
        self.inertial = inertial
        self.collisions = [] if collisions is None else collisions
        self.visuals = [] if visuals is None else visuals
        self.sensors = [] if sensors is None else sensors
        self.projector = projector
        # projector is buggy
        # self.projectors = [] if projectors is None else projectors
        self.audio_sinks = [] if audio_sinks is None else audio_sinks
        self.audio_sources = [] if audio_sources is None else audio_sources
        self.batteries = [] if batteries is None else batteries
        self.lights = [] if lights is None else lights
        self.particle_emitters = [] if particle_emitters is None else particle_emitters

        self._origin.pose = self.pose

        # if projector is not None:
        #     if projector.pose.relative_to is None:
        #         projector.pose.relative_to = name

        # inertial frame is _forced_ to be relative to link
        # if self.inertial is not None:
        #     self.inertial.pose.relative_to = name

        for frame in self._frames:
            if frame.attached_to is None:
                frame.attached_to = self.name
            if frame.pose.relative_to is None:
                frame.pose.relative_to = self.name

        for sensor in sensors:
            if sensor.pose.relative_to is None:
                sensor.pose.relative_to = self.name

            if sensor.camera.pose.relative_to is None:
                sensor.camera.pose.relative_to = f"{self.name}::{sensor.name}"

            for frame in sensor.camera._frames:
                if frame.pose.relative_to is None:
                    frame.pose.relative_to = (
                        f"{self.name}::{sensor.name}::{sensor.camera.name}"
                    )
                if frame.attached_to is None:
                    frame.attached_to = f"{self.name}::{sensor.name}"

            for frame in sensor._frames:
                if frame.pose.relative_to is None:
                    frame.pose.relative_to = f"{self.name}::{sensor.name}"

                if frame.attached_to is None:
                    frame.attached_to = f"{self.name}::{sensor.name}"

    @property
    def origin(self):
        warnings.warn(
            "`Link.origin` is deprecated since SDF v1.7. Use `Link.pose` instead.",
            DeprecationWarning,
        )
        return self._origin

    @property
    def damping(self):
        warnings.warn(
            "`Link.daming` is deprecated since SDF v1.7."
            " Use `Link.velocity_decay` instead.",
            DeprecationWarning,
        )
        return self._damping

    @property
    def frames(self):
        warnings.warn(
            "`Link.frames` is deprecated since SDF v1.7."
            " Use `Model.frames` instead and set `Frame.attached_to` to the name of this link.",
            DeprecationWarning,
        )
        return self._frames

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "Link":
        link_args = {
            "name": specific.name,
            "inertial": specific.inertial,
        }
        args_with_default = {
            "gravity": BoolElement,
            "enable_wind": BoolElement,
            "self_collide": BoolElement,
            "kinematic": BoolElement,
            "origin": Origin,
            "pose": Pose,
            "must_be_base_link": BoolElement,
            "damping": Link.Damping,
            "velocity_decay": Link.VelocityDecay,
            # projector is buggy
            "projector": Projector,
        }
        list_args = {
            "frame": ("frames", Frame),
            "collision": ("collisions", Collision),
            "visual": ("visuals", Visual),
            "sensor": ("sensors", Sensor),
            "audio_sink": ("audio_sinks", AudioSink),
            "audio_source": ("audio_sources", AudioSource),
            "batterie": ("batteries", Battery),
            "light": ("lights", Light),
            "particle_emitter": ("particle_emitters", ParticleEmitter),
        }

        standard_args = cls._prepare_standard_args(
            specific, args_with_default, list_args, version=version
        )
        link_args.update(standard_args)

        return Link(**link_args, sdf_version=version)

    def declared_frames(self) -> Dict[str, tf.Frame]:
        declared_frames = {self.name: tf.Frame(3, name=self.name)}

        for el in chain(self._frames):
            declared_frames.update(el.declared_frames())

        for sensor in self.sensors:
            for name, frame in sensor.declared_frames().items():
                declared_frames[f"{self.name}::{name}"] = frame

        return declared_frames

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
    ) -> tf.Frame:
        parent_name = self.pose.relative_to
        child_name = self.name

        parent = declared_frames[parent_name]
        child = declared_frames[child_name]

        link = self.pose.to_tf_link()
        link(child, parent)

        for frame in self._frames:
            frame.to_static_graph(declared_frames, seed=seed, shape=shape, axis=axis)

        for sensor in self.sensors:
            sensor.to_static_graph(
                declared_frames,
                f"{self.name}::{sensor.name}",
                seed=seed,
                shape=shape,
                axis=axis,
            )

        return declared_frames[self.name]

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame],
    ) -> tf.Frame:
        if self.must_be_base_link:
            parent_name = "world"
            child_name = self.name

            parent = declared_frames[parent_name]
            child = declared_frames[child_name]

            parent_static = _scaffolding[parent_name]
            child_static = _scaffolding[child_name]

            link = tf.CompundLink(parent_static.links_between(child_static))
            link(parent, child)

        for frame in self._frames:
            frame.to_dynamic_graph(
                declared_frames,
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=_scaffolding,
            )

        for sensor in self.sensors:
            parent_name = self.name
            child_name = f"{self.name}::{sensor.name}"

            parent = declared_frames[parent_name]
            child = declared_frames[child_name]

            parent_static = _scaffolding[parent_name]
            child_static = _scaffolding[child_name]

            link = tf.CompundLink(parent_static.links_between(child_static))
            link(parent, child)

            sensor.to_dynamic_graph(
                declared_frames,
                f"{self.name}::{sensor.name}",
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=_scaffolding,
            )

        return declared_frames[self.name]

    class Damping(ElementBase):
        def __init__(
            self, *, linear: float = 0.0, angular: float = 0.0, sdf_version: str
        ) -> None:
            warnings.warn("`Link.Damping` has not been implemented yet.")
            super().__init__(sdf_version=sdf_version)

    class VelocityDecay(ElementBase):
        def __init__(
            self, *, linear: float = 0.0, angular: float = 0.0, sdf_version: str
        ) -> None:
            warnings.warn("`Link.VelocityDecay` has not been implemented yet.")
            super().__init__(sdf_version=sdf_version)



import warnings

from .base import ElementBase


class CollisionEngine(ElementBase):
    def __init__(self, *, sdf_version: str) -> None:
        warnings.warn("`CollisionEngine` has not been implemented yet.")
        super().__init__(sdf_version=sdf_version)



from dataclasses import dataclass, field
from typing import List, Optional
from .joint import Joint
from .link import Link

__NAMESPACE__ = "sdformat/v1.8/actor.xsd"


@dataclass
class Actor:
    """A special kind of model which can have a scripted motion.

    This includes both global waypoint type animations and skeleton
    animations.

    Parameters
    ----------
    skin: Skin file which defines a visual and the underlying skeleton
        which moves it.
    animation: Animation file defines an animation for the skeleton in
        the skin. The skeleton must be compatible with the skin
        skeleton.
    script: Adds scripted trajectories to the actor.
    pose: A position(x,y,z) and orientation(roll, pitch yaw) with
        respect to the frame named in the relative_to attribute.
    link: A physical link with inertia, collision, and visual
        properties. A link must be a child of a model, and any number of
        links may exist in a model.
    joint: A joint connects two links with kinematic and dynamic
        properties. By default, the pose of a joint is expressed in the
        child link frame.
    plugin: A plugin is a dynamically loaded chunk of code. It can exist
        as a child of world, model, and sensor.
    name: A unique name for the actor.
    """

    class Meta:
        name = "actor"

    skin: Optional["Actor.Skin"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    animation: List["Actor.Animation"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    script: Optional["Actor.Script"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    pose: Optional["Actor.Pose"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    link: List[Link] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    joint: List[Joint] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plugin: List["Actor.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Skin:
        """
        Skin file which defines a visual and the underlying skeleton which
        moves it.

        Parameters
        ----------
        filename: Path to skin file, accepted formats: COLLADA, BVH.
        scale: Scale the skin's size.
        """

        filename: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Animation:
        """Animation file defines an animation for the skeleton in the skin.

        The skeleton must be compatible with the skin skeleton.

        Parameters
        ----------
        filename: Path to animation file. Accepted formats: COLLADA,
            BVH.
        scale: Scale for the animation skeleton.
        interpolate_x: Set to true so the animation is interpolated on
            X.
        name: Unique name for animation.
        """

        filename: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        interpolate_x: bool = field(
            default=False,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Script:
        """
        Adds scripted trajectories to the actor.

        Parameters
        ----------
        loop: Set this to true for the script to be repeated in a loop.
            For a fluid continuous motion, make sure the last waypoint
            matches the first one.
        delay_start: This is the time to wait before starting the
            script. If running in a loop, this time will be waited
            before starting each cycle.
        auto_start: Set to true if the animation should start as soon as
            the simulation starts playing. It is useful to set this to
            false if the animation should only start playing only when
            triggered by a plugin, for example.
        trajectory: The trajectory contains a series of keyframes to be
            followed.
        """

        loop: bool = field(
            default=True,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        delay_start: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        auto_start: bool = field(
            default=True,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        trajectory: List["Actor.Script.Trajectory"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Trajectory:
            """
            The trajectory contains a series of keyframes to be followed.

            Parameters
            ----------
            waypoint: Each point in the trajectory.
            id: Unique id for a trajectory.
            type: If it matches the type of an animation, they will be
                played at the same time.
            tension: The tension of the trajectory spline. The default
                value of zero equates to a Catmull-Rom spline, which may
                also cause the animation to overshoot keyframes. A value
                of one will cause the animation to stick to the
                keyframes.
            """

            waypoint: List["Actor.Script.Trajectory.Waypoint"] = field(
                default_factory=list,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            id: Optional[int] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )
            type: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )
            tension: float = field(
                default=0.0,
                metadata={
                    "type": "Attribute",
                },
            )

            @dataclass
            class Waypoint:
                """
                Each point in the trajectory.

                Parameters
                ----------
                time: The time in seconds, counted from the beginning of
                    the script, when the pose should be reached.
                pose: The pose which should be reached at the given
                    time.
                """

                time: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                pose: str = field(
                    default="0 0 0 0 0 0",
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )

    @dataclass
    class Pose:
        """
        Parameters
        ----------
        value:
        relative_to: Name of frame relative to which the pose is
            applied.
        """

        value: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        relative_to: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            },
        )

    @dataclass
    class Plugin:
        """A plugin is a dynamically loaded chunk of code.

        It can exist as a child of world, model, and sensor.

        Parameters
        ----------
        any_element: This is a special element that should not be
            specified in an SDFormat file. It automatically copies child
            elements into the SDFormat element so that a plugin can
            access the data.
        name: A unique name for the plugin, scoped to its parent.
        filename: Name of the shared library to load. If the filename is
            not a full path name, the file will be searched for in the
            configuration paths.
        """

        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )



"""
Inverse Kinematics (IK) Algorithms.

The algorithms in this module assume that the environment is expressed as a
frame graph (see :mod:`skbot.transform`). Once this is done, the problem of
inverse kinematics can be neatly expressed as finding parameters for a set of
:class:`Links <skbot.transform.Link>` such that a point (or set of points,
depending on the algorithm) that has a certain representation in one frame has a
certain representation in another frame.

For example, for a robot like panda we know the tool position in the tool frame
(often 0), and we wish to change the joint angles (joints are
:class:`Links <skbot.transform.Link>`) such that the given position in the tool
frame coincides with a desired position in the world frame.

While using tool and world frame may be most common - after all it is textbook
IK -, this module is not limited to these frames. Any two frames that have (at
least) one valid :func:`transform chain <skbot.transform.Frame.links_between>`
between them that depends on the joints can be used. Expanding on the previous
example, we can specify IK between the camera frame of a static camera and
panda's tool frame.



Functions
---------

.. autosummary::
    :toctree:

    skbot.inverse_kinematics.ccd

"""

from .cyclic_coordinate_descent import ccd

__all__ = ["ccd"]
